import { isU32 } from "../util/number";
import { assert } from "../util/assert";
import { Ins, Del } from "./local-operation";
import { U32Range } from "./u32-range";
/**
 * Abstraction for block insertion and block removal.
 * The abstraction assume a dichotomic traversal of the list.
 * The abstraction can be implemented by an iterator or a recursive data
 * structure.
 *
 * For example, assume we have the list of blocks [A, C, D, E, F] and we insert
 * the block B (A < B < C < D < E < F).
 * B is first compared to D. Because B < D, we attempt to insert in the
 * left "sub-list" [A, C]. C is compared with A then C or C then A, and is
 * finally inserted between A and C.
 */
export class BlockListContext {
    /**
     * Attempt to merge the current block with its predecessor.
     */
    mergeLeft() {
        const curr = this.current();
        const prev = this.predecessor();
        if (prev !== undefined &&
            prev.compare(curr) === -6 /* PREPENDABLE */) {
            this.removePredecessor();
            this.replace(prev.append(curr));
        }
    }
    /**
     * Attempt to merge the current block with its successor.
     */
    mergeRight() {
        const curr = this.current();
        const succ = this.successor();
        if (succ !== undefined &&
            succ.compare(curr) === 6 /* APPENDABLE */) {
            this.removeSuccessor();
            this.replace(curr.append(succ));
        }
    }
    /**
     * Remove current block and attempt to merge its predecessor with
     * its successor.
     */
    removeCurrentMerge() {
        const succ = this.successor();
        if (succ !== undefined) {
            this.replace(succ);
            this.removeSuccessor();
            this.mergeLeft();
        }
        else {
            const prev = this.predecessor();
            if (prev !== undefined) {
                this.replace(prev);
                this.removePredecessor();
            }
            else {
                this.removeCurrent();
            }
        }
    }
    /**
     * @param anchor
     * @param minIndex minimal index in the current sub-list.
     * @return index of `anchor`.
     */
    indexFrom(anchor, minIndex) {
        assert(() => isU32(minIndex), "minIndex ∈ u32");
        const curr = this.current();
        const currIndex = minIndex + this.currentRelativeIndex();
        const currEndIndex = currIndex + curr.length;
        const blockIndex = curr.indexFrom(anchor);
        if (blockIndex === 0) {
            return this.indexFromLeft(anchor, minIndex);
        }
        else if (blockIndex === curr.length) {
            return this.indexFromRight(anchor, currEndIndex);
        }
        else {
            return currIndex + blockIndex;
        }
    }
    /**
     * @param relIndex relative index where the anchor is
     * @param isAfter Is the anchor after `relIndex`?
     * @param f factory
     * @return anchor at `relIndex` in the current list.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    anchorAt(relIndex, isAfter, f) {
        assert(() => isU32(relIndex), "relIndex ∈ u32");
        const curr = this.current();
        const currRelIndex = this.currentRelativeIndex();
        const currRelEndIndex = currRelIndex + curr.length;
        if (relIndex < currRelIndex) {
            return this.anchorAtLeft(relIndex, isAfter, f);
        }
        else if (relIndex === currRelIndex) {
            if (isAfter) {
                return curr.lowerAnchor();
            }
            else {
                return this.anchorAtLeft(relIndex, isAfter, f);
            }
        }
        else if (relIndex < currRelEndIndex) {
            return curr.anchor(relIndex - currRelIndex, !isAfter);
        }
        else if (relIndex === currRelEndIndex && !isAfter) {
            return curr.upperAnchor();
        }
        else {
            const subRelIndex = relIndex - currRelEndIndex;
            return this.anchorAtRight(subRelIndex, isAfter, f);
        }
    }
    /**
     * @param iBlock
     * @return segments of `iBlock` that can be inserted in the current
     *  sub-list.
     */
    insertable(iBlock) {
        const curr = this.current();
        switch (iBlock.compare(curr)) {
            case -6 /* PREPENDABLE */:
            case -7 /* BEFORE */:
                return this.insertableLeft(iBlock);
            case 6 /* APPENDABLE */:
            case 7 /* AFTER */:
                return this.insertableRight(iBlock);
            case -5 /* SPLITTED_BY */: {
                const [lSplit, rSplit] = iBlock.splitWith(curr);
                const lIns = this.insertableLeft(lSplit);
                const rIns = this.insertableRight(rSplit);
                return lIns.concat(rIns);
            }
            case 5 /* SPLITTING */:
                return [iBlock];
            case 3 /* INCLUDING_RIGHT */:
            case -4 /* OVERLAPPING_BEFORE */:
                return this.insertableLeft(iBlock.prependable(curr));
            case 1 /* INCLUDING_LEFT */:
            case 4 /* OVERLAPPING_AFTER */:
                return this.insertableRight(iBlock.appendable(curr));
            case 2 /* INCLUDING_MIDDLE */: {
                const lIns = this.insertableLeft(iBlock.prependable(curr));
                const rIns = this.insertableRight(iBlock.appendable(curr));
                return lIns.concat(rIns);
            }
            case -2 /* INCLUDED_MIDDLE_BY */:
            case -1 /* INCLUDED_RIGHT_BY */:
            case -3 /* INCLUDED_LEFT_BY */:
            case 0 /* EQUAL */:
                return [];
        }
    }
    /**
     * @param iBlock block to insert.
     * @param minIndex minimal index in the current sub-list.
     * @return Performed modifications in terms of local insertions.
     *  The n+1 -th insertion depends on the effect of the n -th insertion.
     */
    insert(iBlock, minIndex) {
        assert(() => isU32(minIndex), "minIndex ∈ u32");
        const curr = this.current();
        const currIndex = minIndex + this.currentRelativeIndex();
        const currEndIndex = currIndex + curr.length;
        switch (iBlock.compare(curr)) {
            case -7 /* BEFORE */:
                return this.insertLeft(iBlock, minIndex);
            case 7 /* AFTER */:
                return this.insertRight(iBlock, currEndIndex);
            case -6 /* PREPENDABLE */: {
                const lIns = this.insertLeft(iBlock, minIndex);
                this.mergeLeft();
                return lIns;
            }
            case 6 /* APPENDABLE */: {
                const rIns = this.insertRight(iBlock, currEndIndex);
                this.mergeRight();
                return rIns;
            }
            case -5 /* SPLITTED_BY */: {
                const [lSplit, rSplit] = iBlock.splitWith(curr);
                const rIns = this.insertRight(rSplit, currEndIndex);
                const lIns = this.insertLeft(lSplit, minIndex);
                return rIns.concat(lIns);
            }
            case 5 /* SPLITTING */: {
                const [lSplit, rSplit] = curr.splitWith(iBlock);
                this.insertPredecessor(lSplit);
                this.replace(iBlock);
                const iBlockIndex = currIndex + lSplit.length;
                this.insertSuccessor(rSplit);
                return [new Ins(iBlockIndex, iBlock.content)];
            }
            case -4 /* OVERLAPPING_BEFORE */:
            case 4 /* OVERLAPPING_AFTER */:
            case -2 /* INCLUDED_MIDDLE_BY */:
            case -1 /* INCLUDED_RIGHT_BY */:
            case -3 /* INCLUDED_LEFT_BY */:
            case 2 /* INCLUDING_MIDDLE */:
            case 3 /* INCLUDING_RIGHT */:
            case 1 /* INCLUDING_LEFT */:
            case 0 /* EQUAL */:
                throw new Error("inserted block intersects with one or more existing blocks");
        }
    }
    /**
     * @param relIndex relative index
     * @param items elements to insert
     * @param factory [mutated] block factory
     * @return block generated with `factory` that contains `items` as content,
     *  and was inserted at `relIndex` in the current sub-list.
     */
    insertAt(relIndex, items, factory) {
        assert(() => isU32(relIndex), "index ∈ u32");
        assert(() => items.length > 0, "items.length > 0");
        const curr = this.current();
        const currRelIndex = this.currentRelativeIndex();
        const currRelEndIndex = currRelIndex + curr.length;
        if (relIndex < currRelIndex) {
            return this.insertAtLeft(relIndex, items, factory);
        }
        else if (relIndex === currRelIndex) {
            const iBlock = factory.between(this.predecessor(), items, curr);
            this.insert(iBlock, 0); // handle merge
            return iBlock;
        }
        else if (relIndex < currRelEndIndex) {
            const [lSplit, rSplit] = curr.splitAt(relIndex - currRelIndex);
            const iBlock = factory.between(lSplit, items, rSplit);
            this.insert(iBlock, 0); // handle split
            return iBlock;
        }
        else if (relIndex === currRelEndIndex) {
            const iBlock = factory.between(curr, items, this.successor());
            this.insert(iBlock, 0); // handle merge
            return iBlock;
        }
        else {
            const subRelIndex = relIndex - currRelEndIndex;
            return this.insertAtRight(subRelIndex, items, factory);
        }
    }
    /**
     * @param dBlock block to remove.
     * @param minIndex minimal index in the current sub-list.
     * @return Performed modifications in terms of local deletions.
     *  The n+1 -th deletion depends on the n -th deletion.
     */
    remove(dBlock, minIndex) {
        assert(() => isU32(minIndex), "minIndex ∈ u32");
        const curr = this.current();
        const currIndex = minIndex + this.currentRelativeIndex();
        const currEndIndex = currIndex + curr.length;
        switch (dBlock.compare(curr)) {
            case -6 /* PREPENDABLE */:
            case -7 /* BEFORE */: {
                const lRmv = this.removeLeft(dBlock, minIndex);
                if (lRmv.length > 0 && lRmv[0].endIndex() === currIndex) {
                    this.mergeLeft();
                }
                return lRmv;
            }
            case 6 /* APPENDABLE */:
            case 7 /* AFTER */: {
                const rRmv = this.removeRight(dBlock, currEndIndex);
                const rRmvLen = rRmv.length;
                if (rRmvLen > 0 && rRmv[rRmvLen - 1].index === currEndIndex) {
                    this.mergeRight();
                }
                return rRmv;
            }
            case 1 /* INCLUDING_LEFT */: {
                const rRmv = this.removeRight(dBlock, currEndIndex);
                const currRmv = this.remove(curr.toLengthBlock(), minIndex); // handle merge
                return rRmv.concat(currRmv);
            }
            case 3 /* INCLUDING_RIGHT */: {
                const lRmv = this.removeLeft(dBlock, minIndex);
                const currRmv = this.remove(curr.toLengthBlock(), minIndex); // handle merge
                return currRmv.concat(lRmv);
            }
            case 2 /* INCLUDING_MIDDLE */: {
                const rRmv = this.removeRight(dBlock, currEndIndex);
                const lRmv = this.removeLeft(dBlock, minIndex);
                const currRmv = this.remove(curr.toLengthBlock(), minIndex); // handle merge
                return rRmv.concat(currRmv).concat(lRmv);
            }
            case -4 /* OVERLAPPING_BEFORE */: {
                const removed = dBlock.intersection(curr);
                const rmv = this.remove(removed, minIndex);
                const lRmv = this.removeLeft(dBlock, minIndex);
                return rmv.concat(lRmv);
            }
            case 4 /* OVERLAPPING_AFTER */: {
                const rRmv = this.removeRight(dBlock, currEndIndex);
                const removed = dBlock.intersection(curr);
                const rmv = this.remove(removed, minIndex);
                return rRmv.concat(rmv);
            }
            case -5 /* SPLITTED_BY */: {
                const [lSplit, rSplit] = dBlock.splitWith(curr);
                const rRmv = this.remove(rSplit, minIndex); // handle merge
                const lRmv = this.remove(lSplit, minIndex); // handle merge
                return rRmv.concat(lRmv);
            }
            case 0 /* EQUAL */: {
                this.removeCurrentMerge();
                return [new Del(currIndex, curr.length)];
            }
            case -3 /* INCLUDED_LEFT_BY */: {
                this.replace(curr.appendable(dBlock));
                return [new Del(currIndex, dBlock.length)];
            }
            case -1 /* INCLUDED_RIGHT_BY */: {
                const replacing = curr.prependable(dBlock);
                this.replace(replacing);
                return [new Del(currIndex + replacing.length, dBlock.length)];
            }
            case -2 /* INCLUDED_MIDDLE_BY */: {
                const replacing = curr.prependable(dBlock);
                this.replace(replacing);
                this.insertSuccessor(curr.appendable(dBlock));
                return [new Del(currIndex + replacing.length, dBlock.length)];
            }
            case 5 /* SPLITTING */:
                return []; // already removed or not already inserted
        }
    }
    /**
     * @param dRange range of indexes to remove
     * @param minIndex minimal index in the current sub-list.
     * @return segments of removed block
     */
    removeAt(dRange, minIndex) {
        assert(() => isU32(minIndex), "minIndex ∈ u32");
        const curr = this.current();
        const currIndex = minIndex + this.currentRelativeIndex();
        const currEndIndex = currIndex + curr.length;
        const currRange = U32Range.fromLength(currIndex, curr.length);
        switch (dRange.compare(currRange)) {
            case -6 /* BEFORE */:
                return this.removeAtLeft(dRange, minIndex);
            case 6 /* AFTER */:
                return this.removeAtRight(dRange, currEndIndex);
            case -5 /* PREPENDABLE */: {
                const lRmv = this.removeAtLeft(dRange, minIndex);
                this.mergeLeft();
                return lRmv;
            }
            case 5 /* APPENDABLE */: {
                const rRmv = this.removeAtRight(dRange, currEndIndex);
                this.mergeRight();
                return rRmv;
            }
            case -4 /* OVERLAPPING_BEFORE */: {
                const rmvRange = dRange.intersection(currRange);
                const currRmv = this.removeAt(rmvRange, minIndex);
                const lRmv = this.removeAtLeft(dRange, minIndex);
                return lRmv.concat(currRmv);
            }
            case 4 /* OVERLAPPING_AFTER */: {
                const rRmv = this.removeAtRight(dRange, currEndIndex);
                const rmvRange = dRange.intersection(currRange);
                const currRmv = this.removeAt(rmvRange, minIndex);
                return currRmv.concat(rRmv);
            }
            case 0 /* EQUAL */: {
                this.removeCurrentMerge();
                return [curr.toLengthBlock()];
            }
            case 1 /* INCLUDING_LEFT */: {
                const rRmv = this.removeAtRight(dRange, currEndIndex);
                this.removeCurrentMerge();
                return [curr.toLengthBlock()].concat(rRmv);
            }
            case 3 /* INCLUDING_RIGHT */: {
                const lRmv = this.removeAtLeft(dRange, minIndex);
                this.removeCurrentMerge();
                return lRmv.concat([curr.toLengthBlock()]);
            }
            case 2 /* INCLUDING_MIDDLE */: {
                const rRmv = this.removeAtRight(dRange, currEndIndex);
                const lRmv = this.removeAtLeft(dRange, minIndex);
                this.removeCurrentMerge();
                return lRmv.concat([curr.toLengthBlock()]).concat(rRmv);
            }
            case -3 /* INCLUDED_LEFT_BY */: {
                const [lSplit, rSplit] = curr.splitAt(dRange.length);
                this.replace(rSplit);
                return [lSplit.toLengthBlock()];
            }
            case -1 /* INCLUDED_RIGHT_BY */: {
                const splitIndex = dRange.lower - currRange.lower;
                const [lSplit, rSplit] = curr.splitAt(splitIndex);
                this.replace(lSplit);
                return [rSplit.toLengthBlock()];
            }
            case -2 /* INCLUDED_MIDDLE_BY */: {
                const rSplitIndex = dRange.upper() + 1 - currRange.lower;
                const [lSplit, rSplit] = curr.splitAt(rSplitIndex);
                const lSplitIndex = dRange.lower - currRange.lower;
                const [llSplit, lrSplit] = lSplit.splitAt(lSplitIndex);
                this.replace(llSplit);
                this.insertSuccessor(rSplit);
                return [lrSplit.toLengthBlock()];
            }
        }
    }
}
//# sourceMappingURL=block-list-context.js.map