import { BlockListContext } from "./block-list-context";
import { assert } from "../util/assert";
import { isU32 } from "../util/number";
import { isObject } from "../util/data-validation";
export class DeltaReplicatedList {
    constructor(list, vv) {
        this.list = list;
        this.versionVector = vv;
    }
    static from(list) {
        assert(() => list.length === 0, "list must be empty");
        return new DeltaReplicatedList(list, Object.create(null));
    }
    /**
     * @param opFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain(opFromPlain) {
        return (x) => {
            if (isObject(x) &&
                isObject(x.versionVector)) {
                const versionVector = Object.create(null);
                for (const replica in x.versionVector) {
                    if (x.versionVector.hasOwnProperty(replica)) {
                        const seq = x.versionVector[replica];
                        if (isU32(seq)) {
                            versionVector[replica] = seq;
                        }
                    }
                }
                const list = opFromPlain(x.list);
                if (list !== undefined) {
                    return new DeltaReplicatedList(list, versionVector);
                }
            }
            return undefined;
        };
    }
    readonlyVersionvector() {
        return this.versionVector;
    }
    get length() {
        return BlockListContext.length;
    }
    reduceBlock(f, prefix) {
        return this.list.reduceBlock(f, prefix);
    }
    /**
     * @param prefix
     * @return Concatenated version prefixed by `prefix`.
     */
    concatenated(prefix) {
        return this.list.concatenated(prefix);
    }
    /**
     * Non-cryptographic way to approximate object identity.
     * Do not take the blocks' content into account.
     */
    structuralHashCode() {
        return this.list.structuralHashCode();
    }
    lastSeq(replica) {
        const maybeLastSeq = this.versionVector[replica];
        if (maybeLastSeq !== undefined) {
            return maybeLastSeq;
        }
        return 0;
    }
    /**
     * @param anchor
     * @return index of `anchor`.
     */
    indexFrom(anchor) {
        return this.list.indexFrom(anchor);
    }
    /**
     * Complexity: O(1)
     *
     * @param delta
     * @return parts of `delta` that can be inserted.
     */
    insertable(delta) {
        const lastSeq = this.lastSeq(String(delta.replica()));
        const deltaSeqs = delta.seqs();
        if (deltaSeqs.lower > lastSeq) {
            return [delta];
        }
        else if (deltaSeqs.upper() > lastSeq) {
            const remaining = delta.rightSplitAt(lastSeq + 1 - deltaSeqs.lower);
            return [remaining];
        }
        return [];
    }
    /**
     * Complexity: O(S) where S depends of the chosen implementation.
     * e.g. for a balanced tree: S = log2(N) with N the number of blocks.
     *
     * @param delta
     * @return parts of `delta` that was removed.
     */
    removed(delta) {
        const lastSeq = this.lastSeq(String(delta.replica()));
        const deltaSeqs = delta.seqs();
        if (deltaSeqs.lower > lastSeq) {
            return [];
        }
        else if (deltaSeqs.upper() > lastSeq) {
            const splittingIndex = lastSeq + 1 - deltaSeqs.lower;
            const remaining = delta.leftSplitAt(splittingIndex);
            return this.list.insertable(remaining).map((block) => {
                return block.toLengthBlock();
            });
        }
        return this.list.insertable(delta).map((block) => {
            return block.toLengthBlock();
        });
    }
    bumpVersion(delta) {
        const lastSeq = this.lastSeq(String(delta.replica()));
        const deltaSeqs = delta.seqs();
        if (deltaSeqs.lower > lastSeq + 1) {
            throw new Error("FIFO violation");
        }
        else {
            this.versionVector[delta.replica()] = Math.max(lastSeq, deltaSeqs.upper());
        }
    }
    insert(delta) {
        let result = [];
        for (const d of this.insertable(delta)) {
            const ins = this.list.insert(d);
            result = result.concat(ins);
        }
        this.bumpVersion(delta);
        return result;
    }
    remove(delta) {
        this.bumpVersion(delta);
        return this.list.remove(delta);
    }
    quickFilter(delta) {
        if (delta.isLengthBlock()) {
            return [delta];
        }
        else {
            return this.insertable(delta);
        }
    }
    applyDelta(delta) {
        if (delta.isLengthBlock()) {
            return this.remove(delta);
        }
        else {
            return this.insert(delta);
        }
    }
    /**
     * [Mutation]
     * Unidirectional merge from other to this.
     * This operation is idempotent.
     *
     * @param other
     * @return Performed modifications in terms of local operations.
     *  The n+1 -th operation depends on the n -th operation.
     */
    merge(other) {
        // determine inserted blocks in `other` and missing in `this`
        const inserted = other.reduceBlock((acc, block) => {
            return acc.concat(this.insertable(block));
        }, []);
        // determine removed block in `other` and present in `this`
        const removed = this.reduceBlock((acc, block) => {
            return acc.concat(other.removed(block));
        }, []);
        // apply raw insertions and raw removals (bypassing causal context)
        let result = [];
        for (const iBlock of inserted) {
            result = result.concat(this.list.insert(iBlock));
        }
        for (const dBlock of removed) {
            result = result.concat(this.list.remove(dBlock));
        }
        // merge causal contexts
        for (const replica in other.readonlyVersionvector()) {
            const otherLastSeq = other.readonlyVersionvector()[replica];
            const lastSeq = this.versionVector[replica];
            if (lastSeq !== undefined && otherLastSeq !== undefined) {
                this.versionVector[replica] = Math.max(lastSeq, otherLastSeq);
            }
            else if (otherLastSeq !== undefined) {
                this.versionVector[replica] = otherLastSeq;
            }
        }
        return result;
    }
}
export class DeltaEditableReplicatedList extends DeltaReplicatedList {
    constructor(list, vv) {
        super(list, vv);
        this.list = list;
    }
    static from(list) {
        assert(() => list.length === 0, "list must be empty");
        return new DeltaEditableReplicatedList(list, Object.create(null));
    }
    /**
     * @param opFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain(opFromPlain) {
        return (x) => {
            if (isObject(x) &&
                isObject(x.versionVector)) {
                const versionVector = Object.create(null);
                for (const replica in x.versionVector) {
                    if (x.versionVector.hasOwnProperty(replica)) {
                        const seq = x.versionVector[replica];
                        if (isU32(seq)) {
                            versionVector[replica] = seq;
                        }
                    }
                }
                const list = opFromPlain(x.list);
                if (list !== undefined) {
                    return new DeltaEditableReplicatedList(list, versionVector);
                }
            }
            return undefined;
        };
    }
    /**
     * @param index index where the anchor is
     * @param isAfter Is the anchor after `index`?
     * @return anchor at `index`.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    anchorAt(index, isAfter) {
        assert(() => isU32(index), "index ∈ u32");
        return this.list.anchorAt(index, isAfter);
    }
    /**
     * [Mutation]
     * Insert {@link items } at {@link index}.
     *
     * @param index 0-based index; Where to insert.
     * @param items elements to insert.
     * @return Delta which represents the insertion.
     */
    insertAt(index, items) {
        assert(() => isU32(index), "index ∈ u32");
        assert(() => items.length > 0, "items is not empty");
        const iBlock = this.list.insertAt(index, items);
        this.versionVector[iBlock.replica()] = iBlock.seqs().upper();
        return iBlock;
    }
    /**
     * [Mutation]
     * Remove a number of {@link length } elements from {@link index }
     *
     * @param index 0-based index.
     * @param length Number of elements to remove.
     * @return Delta which represents the deletion.
     */
    removeAt(index, length) {
        assert(() => isU32(index), "index ∈ u32");
        assert(() => isU32(length), "length ∈ u32");
        return this.list.removeAt(index, length);
    }
}
//# sourceMappingURL=delta-replicated-list.js.map