/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
import { assert, heavyAssert } from "../util/assert";
import { isU32 } from "../util/number";
import { isObject } from "../util/data-validation";
/**
 * Interval of u32.
 */
export class U32Range {
    /**
     * @param lower Lower bound.
     * @param length Number of elements in the range.
     */
    constructor(lower, length) {
        this.lower = lower;
        this.length = length;
    }
    /**
     * @param lower {@link IntInterval#lower }
     * @param length {@link IntInterval#length }
     * @return New range.
     */
    static fromLength(lower, length) {
        assert(() => isU32(lower), "lower ∈ u32");
        assert(() => isU32(length), "length ∈ u32");
        assert(() => length > 0, "length > 0");
        //assert(() => isu32(lower + length))
        return new U32Range(lower, length);
    }
    /**
     * @param lower {@link IntInterval#lower }
     * @param upper {@link IntInterval#length }
     * @return New range.
     */
    static fromBounds(lower, upper) {
        assert(() => isU32(lower), "lower ∈ u32");
        assert(() => isU32(upper), "upper ∈ u32");
        assert(() => lower <= upper, "lower <= upper");
        return U32Range.fromLength(lower, upper - lower + 1);
    }
    /**
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    static fromPlain(x) {
        if (isObject(x) &&
            isU32(x.lower) &&
            isU32(x.length) &&
            x.length > 0) {
            return new U32Range(x.lower, x.length);
        }
        return undefined;
    }
    // Access
    /**
     * @example
     * nth(0) == lower
     * nth(length - 1) == upper
     *
     * @param nth 0-based
     * @return n-th element.
     */
    nth(nth) {
        assert(() => isU32(nth), "nth ∈ u32");
        assert(() => nth < this.length, "valid nth");
        assert(() => isU32(this.lower + nth), "No integer overflow");
        return this.lower + nth;
    }
    /**
     * Upper bound.
     */
    upper() {
        return this.nth(this.length - 1);
    }
    // Derivation
    /**
     * @param other
     * @return {@link other } appended to this.
     */
    append(other) {
        heavyAssert(() => other.compare(this) === 5 /* APPENDABLE */, "other is appendable to this");
        return new U32Range(this.lower, this.length + other.length);
    }
    /**
     * @param other
     *      this includes or overlaps after {@link other }
     * @return Part of this block which can be append to {@link block }.
     */
    appendable(other) {
        heavyAssert(() => this.hasAppendable(other), "this has an appendable segment to other");
        return U32Range.fromLength(other.upper() + 1, this.upper());
    }
    /**
     * @param other
     *      this includes or overlaps before {@link other }
     * @return Part of this block which can be prepend to {@link block }.
     */
    prependable(other) {
        heavyAssert(() => this.hasPrependable(other), "this has a prependable segment to other");
        return U32Range.fromLength(this.lower, other.upper() - this.lower);
    }
    /**
     * @param other
     *      this and {@link other } intersect.
     * @return Intersection part between this and {@link other }.
     */
    intersection(other) {
        heavyAssert(() => this.hasIntersection(other), "this intersects other.");
        return U32Range.fromBounds(Math.max(this.lower, other.lower), Math.min(this.upper(), other.upper()));
    }
    // Status
    /**
     * @param other
     * @return Are this and {@link other } an intersection?
     */
    hasIntersection(other) {
        const cmp = this.compare(other);
        return (-4 /* OVERLAPPING_BEFORE */ <= cmp &&
            cmp <= 4 /* OVERLAPPING_AFTER */);
    }
    /**
     * @param other
     * @return Has this an appendable segment to {@link other }?
     */
    hasAppendable(other) {
        const cmp = this.compare(other);
        return (cmp === 4 /* OVERLAPPING_AFTER */ ||
            cmp === 1 /* INCLUDING_LEFT */ ||
            cmp === 2 /* INCLUDING_MIDDLE */);
    }
    /**
     * @param other
     * @return Has this a prependable segment to {@link other }?
     */
    hasPrependable(other) {
        const cmp = this.compare(other);
        return (cmp === -4 /* OVERLAPPING_BEFORE */ ||
            cmp === 3 /* INCLUDING_RIGHT */ ||
            cmp === 2 /* INCLUDING_MIDDLE */);
    }
    /**
     * @param other
     * @return this [order relation] other.
     */
    compare(other) {
        const thisUpper = this.upper();
        const otherUpper = other.upper();
        if (thisUpper < other.lower) {
            if (thisUpper + 1 === other.lower) {
                return -5 /* PREPENDABLE */;
            }
            else {
                return -6 /* BEFORE */;
            }
        }
        else if (otherUpper < this.lower) {
            if (otherUpper + 1 === this.lower) {
                return 5 /* APPENDABLE */;
            }
            else {
                return 6 /* AFTER */;
            }
        }
        else {
            if (this.lower === other.lower) {
                if (thisUpper === otherUpper) {
                    return 0 /* EQUAL */;
                }
                else if (thisUpper < otherUpper) {
                    return -3 /* INCLUDED_LEFT_BY */;
                }
                else {
                    return 1 /* INCLUDING_LEFT */;
                }
            }
            else if (this.lower < other.lower) {
                if (otherUpper === thisUpper) {
                    return 3 /* INCLUDING_RIGHT */;
                }
                else if (otherUpper < thisUpper) {
                    return 2 /* INCLUDING_MIDDLE */;
                }
                else {
                    return -4 /* OVERLAPPING_BEFORE */;
                }
            }
            else {
                if (thisUpper === otherUpper) {
                    return -1 /* INCLUDED_RIGHT_BY */;
                }
                else if (thisUpper < otherUpper) {
                    return -2 /* INCLUDED_MIDDLE_BY */;
                }
                else {
                    return 4 /* OVERLAPPING_AFTER */;
                }
            }
        }
    }
}
//# sourceMappingURL=u32-range.js.map