/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
import { assert } from "../../util/assert";
import { isObject } from "../../util/data-validation";
import { compareU32, hashCodeOf, isU32, U32_BOTTOM, U32_TOP, } from "../../util/number";
import { lexCompareOrdering } from "../../util/ordering";
/**
 * @final this class cannot be extended
 *
 * Part of a {@link SimpleDotPos}.
 *
 * A part is a lexicographic triplet (priority, replica, seq).
 * The dot (replica, seq) uniquely identifies.
 */
export class SimpleDotPosPart {
    /**
     * @param priority {@link SimpleDotPosPart#priority }
     * @param replica {@link SimpleDotPosPart#replica }
     * @param seq {@link SimpleDotPosPart#seq }
     */
    constructor(priority, replica, seq) {
        assert(() => isU32(priority), "random ∈ u32");
        assert(() => isU32(replica), "replica ∈ u32");
        assert(() => isU32(seq), "seq ∈ u32");
        this.priority = priority;
        this.replica = replica;
        this.seq = seq;
    }
    /**
     * @param priority {@link SimpleDotPosPart#priority }
     *      ∉ {U32_BOTTOM, U32_TOP}
     * @param replica {@link SimpleDotPosPart#replica }
     *      !== U32_TOP
     * @param seq {@link SimpleDotPosPart#seq }
     */
    static from(priority, replica, seq) {
        assert(() => priority !== U32_BOTTOM && priority !== U32_TOP, "priority ∉ {U32_BOTTOM, U32_TOP}");
        assert(() => replica !== U32_TOP, "replica != U32_TOP. Reserved for BOTTOM and TOP.");
        assert(() => seq !== 0, "seq != 0. Reserved for convenience.");
        return new SimpleDotPosPart(priority, replica, seq);
    }
    /**
     * @note {@link SimpleDotPosPart#BOTTOM} and {@link SimpleDotPosPart#TOP}
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    static fromPlain(x) {
        if (isObject(x) &&
            isU32(x.priority) &&
            isU32(x.replica) &&
            isU32(x.seq) &&
            x.seq !== 0) {
            return new SimpleDotPosPart(x.priority, x.replica, x.seq);
        }
        return undefined;
    }
    // Derivation
    /**
     * @param seq The seq of the new SimpleDotPosPart
     * @return part with the same base,
     *  but with a different seq
     */
    withSeq(seq) {
        assert(() => isU32(seq), "seq ∈ u32");
        return new SimpleDotPosPart(this.priority, this.replica, seq);
    }
    /**
     * @return [priority, replica, seq]
     */
    asTuple() {
        return [this.priority, this.replica, this.seq];
    }
    /**
     * Non-cryptographic way to approximate object identity.
     */
    hashCode() {
        return hashCodeOf(this.asTuple());
    }
    // Status
    /**
     * @example
     * SimpleDotPosPart(0, 0, _) == SimpleDotPosPart(0, 0, _)
     * SimpleDotPosPart(0, 0, _) < SimpleDotPosPart(1, 0, _)
     * SimpleDotPosPart(1, 0, _) < SimpleDotPosPart(1, 1, _)
     *
     * Lexicographic order of pair (priority, replica).
     *
     * @param other
     * @return this [Order relation] {@link other }, regardless their seqs
     *  ({@link SimpleDotPosPart#seq}).
     */
    compareBase(other) {
        return lexCompareOrdering(compareU32(this.priority, other.priority), compareU32(this.replica, other.replica));
    }
    /**
     * @example
     * SimpleDotPosPart(0, 0, 0) < SimpleDotPosPart(1, 0, 0)
     * SimpleDotPosPart(1, 0, 0) < SimpleDotPosPart(1, 1, 0)
     * SimpleDotPosPart(1, 1, 0) < SimpleDotPosPart(1, 1, 1)
     *
     * Lexicographic order of triple (priority, replica, seq).
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    compare(other) {
        if (this === other) {
            return 0 /* EQUAL */;
        }
        else {
            return lexCompareOrdering(this.compareBase(other), compareU32(this.seq, other.seq));
        }
    }
}
/**
 * Lowest part.
 * Here the pair (replica, seq) has no meaning.
 */
SimpleDotPosPart.BOTTOM = new SimpleDotPosPart(U32_BOTTOM, U32_TOP, 1);
/**
 * Greatest part.
 * Here the pair (replica, seq) has no meaning.
 */
SimpleDotPosPart.TOP = new SimpleDotPosPart(U32_TOP, U32_TOP, 2);
//# sourceMappingURL=simple-dot-pos-part.js.map