import { Ins } from "../../core/local-operation";
import { heavyAssert } from "../../util/assert";
import { BlockListContext } from "../../core/block-list-context";
import { isObject } from "../../util/data-validation";
export function lengthOf(node) {
    if (node !== undefined) {
        return node.length;
    }
    return 0;
}
function rankOf(node) {
    if (node !== undefined) {
        return node.rank;
    }
    return 0;
}
export class ValuedNode extends BlockListContext {
    constructor(l, block, r) {
        super();
        this.left = l;
        this.block = block;
        this.right = r;
        this.isSelfRemoved = false;
        this.length = 0;
        this.rank = 0;
        this.update();
    }
    static leaf(block) {
        return new ValuedNode(undefined, block, undefined);
    }
    /**
     * @param blockFromPlain
     * @return function that accepts a value and returns a Node from
     *  the value, or undefined if the value is mal-formed
     */
    static fromPlain(blockFromPlain) {
        return (x) => {
            if (isObject(x)) {
                const left = ValuedNode.fromPlain(blockFromPlain)(x.left);
                const right = ValuedNode.fromPlain(blockFromPlain)(x.right);
                const block = blockFromPlain(x.block);
                if (block !== undefined) {
                    return new ValuedNode(left, block, right);
                }
            }
            return undefined;
        };
    }
    // Basic changes
    setLeft(l) {
        ;
        this.left = l;
        this.update();
    }
    replace(block) {
        ;
        this.block = block;
        this.update();
    }
    setRight(r) {
        ;
        this.right = r;
        this.update();
    }
    update() {
        this.length =
            lengthOf(this.left) + this.block.length + lengthOf(this.right);
        this.rank = 1 + Math.max(rankOf(this.left), rankOf(this.right));
    }
    // Balancing
    balanceFactor() {
        return rankOf(this.right) - rankOf(this.left);
    }
    isBalanced() {
        const balanceFactor = this.balanceFactor();
        return -2 < balanceFactor && balanceFactor < 2;
    }
    isRightUnbalanced() {
        return this.balanceFactor() >= 2;
    }
    isRightOriented() {
        return this.balanceFactor() === 1;
    }
    isLeftOriented() {
        return this.balanceFactor() <= -2;
    }
    isSlighlyLeftUnbalanced() {
        return this.balanceFactor() === -1;
    }
    rotateLeft() {
        const r = this.right;
        this.setRight(r.left);
        r.setLeft(this);
        return r;
    }
    rotateRight() {
        const l = this.left;
        this.setLeft(l.right);
        l.setRight(this);
        return l;
    }
    balance() {
        let self = this;
        while (self.isRightUnbalanced()) {
            if (self.right.isSlighlyLeftUnbalanced()) {
                self.setRight(self.right.rotateRight());
            }
            self = self.rotateLeft();
        }
        while (self.isLeftOriented()) {
            if (self.left.isRightOriented()) {
                self.setLeft(self.left.rotateLeft());
            }
            self = self.rotateRight();
        }
        return self;
    }
    // Self removal
    updated() {
        if (this.isSelfRemoved) {
            return undefined;
        }
        return this.balance();
    }
    // Traversal
    reduceBlock(f, prefix) {
        let leftPrefix = prefix;
        if (this.left !== undefined) {
            leftPrefix = this.left.reduceBlock(f, prefix);
        }
        leftPrefix = f(leftPrefix, this.block);
        if (this.right !== undefined) {
            return this.right.reduceBlock(f, leftPrefix);
        }
        return leftPrefix;
    }
    // Impl
    current() {
        return this.block;
    }
    currentRelativeIndex() {
        return lengthOf(this.left);
    }
    min() {
        if (this.left !== undefined) {
            return this.left.min(); // tail recursion
        }
        return this.block;
    }
    max() {
        if (this.right !== undefined) {
            return this.right.max(); // tail recursion
        }
        return this.block;
    }
    predecessor() {
        if (this.left !== undefined) {
            return this.left.max();
        }
        return undefined;
    }
    successor() {
        if (this.right !== undefined) {
            return this.right.min();
        }
        return undefined;
    }
    insertMin(iBlock) {
        if (this.left !== undefined) {
            this.setLeft(this.left.insertMin(iBlock));
        }
        else {
            heavyAssert(() => this.block.compare(iBlock) === 7 /* AFTER */, "iBlock < block");
            this.setLeft(ValuedNode.leaf(iBlock));
        }
        return this.balance();
    }
    insertMax(iBlock) {
        if (this.right !== undefined) {
            this.setRight(this.right.insertMax(iBlock));
        }
        else {
            heavyAssert(() => this.block.compare(iBlock) === -7 /* BEFORE */, "block < iBlock");
            this.setRight(ValuedNode.leaf(iBlock));
        }
        return this.balance();
    }
    // Anchor
    /** @Override */
    indexFromLeft(anchor, minIndex) {
        if (this.left !== undefined) {
            return this.left.indexFrom(anchor, minIndex);
        }
        return minIndex;
    }
    /** @Override */
    indexFromRight(anchor, minIndex) {
        if (this.right !== undefined) {
            return this.right.indexFrom(anchor, minIndex);
        }
        return minIndex;
    }
    /** @Override */
    anchorAtLeft(relIndex, isAfter, f) {
        if (this.left !== undefined) {
            return this.left.anchorAt(relIndex, isAfter, f);
        }
        return f.bottomAnchor;
    }
    /** @Override */
    anchorAtRight(relIndex, isAfter, f) {
        if (this.right !== undefined) {
            return this.right.anchorAt(relIndex, isAfter, f);
        }
        else if (isAfter) {
            return f.topAnchor;
        }
        else {
            return this.block.upperAnchor();
        }
    }
    // Insertion
    /** @Override */
    insertableLeft(iBlock) {
        if (this.left !== undefined) {
            return this.left.insertable(iBlock);
        }
        return [iBlock];
    }
    /** @Override */
    insertableRight(iBlock) {
        if (this.right !== undefined) {
            return this.right.insertable(iBlock);
        }
        return [iBlock];
    }
    /** @Override */
    insertPredecessor(iBlock) {
        if (this.left !== undefined) {
            this.setLeft(this.left.insertMax(iBlock));
        }
        else {
            this.setLeft(ValuedNode.leaf(iBlock));
        }
        // delegate balance to parent
    }
    /** @Override */
    insertSuccessor(iBlock) {
        if (this.right !== undefined) {
            this.setRight(this.right.insertMin(iBlock));
        }
        else {
            this.setRight(ValuedNode.leaf(iBlock));
        }
        // delegate balance to parent
    }
    /** @Override */
    insertLeft(iBlock, minIndex) {
        let ins;
        if (this.left !== undefined) {
            ins = this.left.insert(iBlock, minIndex);
            this.setLeft(this.left.balance());
        }
        else {
            this.setLeft(ValuedNode.leaf(iBlock));
            ins = [new Ins(minIndex, iBlock.content)];
        }
        // delegate balance to parent
        return ins;
    }
    /** @Override */
    insertRight(iBlock, minIndex) {
        let ins;
        if (this.right !== undefined) {
            ins = this.right.insert(iBlock, minIndex);
            this.setRight(this.right.balance());
        }
        else {
            // should not happen
            this.setRight(ValuedNode.leaf(iBlock));
            ins = [new Ins(minIndex, iBlock.content)];
        }
        // delegate balance to parent
        return ins;
    }
    /** @Override */
    insertAtLeft(idx, items, f) {
        let ins;
        if (this.left !== undefined) {
            ins = this.left.insertAt(idx, items, f);
            this.setLeft(this.left.balance());
        }
        else {
            // should not happen
            ins = this.insertAt(0, items, f);
        }
        // delegate balance to parent
        return ins;
    }
    /** @Override */
    insertAtRight(idx, items, f) {
        let ins;
        if (this.right !== undefined) {
            ins = this.right.insertAt(idx, items, f);
            this.setRight(this.right.balance());
        }
        else {
            const lastIdx = this.currentRelativeIndex() + this.block.length;
            ins = this.insertAt(lastIdx, items, f);
        }
        // delegate balance to parent
        return ins;
    }
    // Removal
    /** @Override */
    removeCurrent() {
        this.isSelfRemoved = true;
    }
    removeMin() {
        if (this.left !== undefined) {
            this.setLeft(this.left.removeMin());
            return this.balance();
        }
        return this.right;
    }
    removeMax() {
        if (this.right !== undefined) {
            this.setRight(this.right.removeMax());
            return this.balance();
        }
        return this.left;
    }
    /** @Override */
    removePredecessor() {
        if (this.left !== undefined) {
            this.setLeft(this.left.removeMax());
        }
        // delegate balance to parent
    }
    /** @Override */
    removeSuccessor() {
        if (this.right !== undefined) {
            this.setRight(this.right.removeMin());
        }
        // delegate balance to parent
    }
    /** @Override */
    removeLeft(dBlock, minIndex) {
        if (this.left !== undefined) {
            const rmv = this.left.remove(dBlock, minIndex);
            this.setLeft(this.left.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    }
    /** @Override */
    removeRight(dBlock, minIndex) {
        if (this.right !== undefined) {
            const rmv = this.right.remove(dBlock, minIndex);
            this.setRight(this.right.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    }
    /** @Override */
    removeAtLeft(dRange, minIndex) {
        if (this.left !== undefined) {
            const rmv = this.left.removeAt(dRange, minIndex);
            this.setLeft(this.left.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    }
    /** @Override */
    removeAtRight(dRange, minIndex) {
        if (this.right !== undefined) {
            const rmv = this.right.removeAt(dRange, minIndex);
            this.setRight(this.right.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    }
}
//# sourceMappingURL=avl-list-node.js.map