import { OpReplicatedList, } from "../../core/op-replicated-list";
import { ValuedNode, lengthOf } from "./avl-list-node";
import { isU32 } from "../../util/number";
import { Ins } from "../../core/local-operation";
import { assert } from "../../util/assert";
import { U32Range } from "../../core/u32-range";
import { isObject } from "../../util/data-validation";
/**
 * An {@see OpReplicatedList } that uses an AVL tree.
 */
export class OpAvlList extends OpReplicatedList {
    /**
     * @param root {@see EditableOpAvlList#root }
     */
    constructor(root) {
        super();
        this.root = root;
    }
    /**
     * @param v empty value of type E
     * @return new empty list.
     */
    static empty() {
        return new OpAvlList(undefined);
    }
    /**
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain(f, itemsFromPlain) {
        return (x) => {
            if (isObject(x)) {
                const blockFromPlain = f.blockFromPlain(itemsFromPlain);
                const root = ValuedNode.fromPlain(blockFromPlain)(x.root);
                return new OpAvlList(root);
            }
            return undefined;
        };
    }
    /** @Override */
    get length() {
        return lengthOf(this.root);
    }
    /** @Override */
    indexFrom(anchor) {
        if (this.root !== undefined) {
            return this.root.indexFrom(anchor, 0);
        }
        else {
            return 0;
        }
    }
    /** @Override */
    reduceBlock(f, prefix) {
        if (this.root !== undefined) {
            return this.root.reduceBlock(f, prefix);
        }
        return prefix;
    }
    /** @Override */
    insertable(delta) {
        if (this.root !== undefined) {
            return this.root.insertable(delta);
        }
        return [];
    }
    // Modification
    /** @Override */
    insert(delta) {
        if (this.root !== undefined) {
            const ins = this.root.insert(delta, 0);
            this.root = this.root.balance();
            return ins;
        }
        else {
            this.root = ValuedNode.leaf(delta);
            return [new Ins(0, delta.content)];
        }
    }
    /** @Override */
    remove(delta) {
        if (this.root !== undefined) {
            const rmv = this.root.remove(delta, 0);
            this.root = this.root.updated();
            return rmv;
        }
        return [];
    }
}
/**
 * An {@see EditableOpReplicatedList } that uses an AVL tree.
 */
export class EditableOpAvlList extends OpAvlList {
    /**
     * @param root {@see EditableOpAvlList#root }
     * @param factory {@see EditableOpAvlList#factory }
     */
    constructor(root, factory) {
        super(root);
        this.factory = factory;
    }
    /**
     * @param factory factory of blocks
     * @param v empty value of type E (only used for type inference)
     * @return new empty list.
     */
    static emptyWith(factory) {
        return new EditableOpAvlList(undefined, factory);
    }
    /**
     *
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain(f, itemsFromPlain) {
        return (x) => {
            if (isObject(x)) {
                const factory = f.fromPlain(x.factory);
                if (factory !== undefined) {
                    const blockFromPlain = f.blockFromPlain(itemsFromPlain);
                    const root = ValuedNode.fromPlain(blockFromPlain)(x.root);
                    return new EditableOpAvlList(root, factory);
                }
            }
            return undefined;
        };
    }
    /** @Override */
    anchorAt(index, isAfter) {
        assert(() => isU32(index), "index ∈ u32");
        if (this.root !== undefined) {
            return this.root.anchorAt(index, isAfter, this.factory);
        }
        else if (isAfter) {
            return this.factory.topAnchor;
        }
        else {
            return this.factory.bottomAnchor;
        }
    }
    /** @Override */
    remove(delta) {
        this.factory.garbageCollect(delta);
        return super.remove(delta);
    }
    /** @Override */
    insertAt(index, items) {
        assert(() => isU32(index), "index ∈ u32");
        assert(() => index <= this.length, "valid index");
        assert(() => items.length > 0, "items is not empty");
        if (this.root !== undefined) {
            const iBlock = this.root.insertAt(index, items, this.factory);
            this.root = this.root.balance();
            return iBlock;
        }
        else {
            const iBlock = this.factory.from(items);
            this.root = ValuedNode.leaf(iBlock);
            return iBlock;
        }
    }
    /** @Override */
    removeAt(index, length) {
        assert(() => isU32(index), "index ∈ u32");
        assert(() => index < this.length, "valid index");
        assert(() => isU32(length), "length ∈ u32");
        assert(() => index + length <= this.length, "valid end index");
        if (this.root !== undefined) {
            const indexRange = U32Range.fromLength(index, length);
            const rmv = this.root.removeAt(indexRange, 0);
            this.root = this.root.updated();
            if (this.factory.canGarbageCollect()) {
                for (const dBlock of rmv) {
                    this.factory.garbageCollect(dBlock);
                }
            }
            return rmv;
        }
        return [];
    }
}
//# sourceMappingURL=op-avl-list.js.map