/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
import { assert } from "../../util/assert";
import { isU32 } from "../../util/number";
import { Sentinel } from "./linked-list-cell";
import { OpReplicatedList, } from "../../core/op-replicated-list";
import { isObject } from "../../util/data-validation";
/**
 * An {@see OpReplicatedList } that uses an AVL tree.
 */
export class OpLinkedList extends OpReplicatedList {
    /**
     * New empty list.
     */
    constructor(root, length) {
        super();
        this.root = root;
        this.length = length;
    }
    /**
     * @param v empty value of type E
     * @return new empty list.
     */
    static empty() {
        return new OpLinkedList(new Sentinel(), 0);
    }
    /**
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain(f, itemsFromPlain) {
        return (x) => {
            if (isObject(x) &&
                isU32(x.length)) {
                const blockFromPlain = f.blockFromPlain(itemsFromPlain);
                const root = Sentinel.fromPlain(blockFromPlain)(x.root);
                if (root !== undefined) {
                    return new OpLinkedList(root, x.length);
                }
            }
            return undefined;
        };
    }
    /** @Override */
    reduceBlock(f, prefix) {
        if (this.root !== undefined) {
            return this.root.reduceBlock(f, prefix);
        }
        return prefix;
    }
    /** @Override */
    indexFrom(anchor) {
        return this.root.indexFrom(anchor, 0);
    }
    /** @Override */
    insertable(delta) {
        return this.root.insertable(delta);
    }
    // Modification
    /** @Override */
    insert(block) {
        const result = this.root.insert(block, 0);
        this.length = result.reduce((acc, v) => acc + v.length, this.length);
        return result;
    }
    /** @Override */
    remove(block) {
        const result = this.root.remove(block, 0);
        this.length = result.reduce((acc, v) => acc - v.length, this.length);
        return result;
    }
}
/**
 * An {@see EditableOpReplicatedList } that uses an AVL tree.
 */
export class EditableOpLinkedList extends OpLinkedList {
    /**
     * @param factory strategy of block generation.
     */
    constructor(root, length, factory) {
        super(root, length);
        this.factory = factory;
    }
    /**
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain(f, itemsFromPlain) {
        return (x) => {
            if (isObject(x) &&
                isU32(x.length)) {
                const blockFromPlain = f.blockFromPlain(itemsFromPlain);
                const factory = f.fromPlain(x.factory);
                const root = Sentinel.fromPlain(blockFromPlain)(x.root);
                if (root !== undefined && factory !== undefined) {
                    return new EditableOpLinkedList(root, x.length, factory);
                }
            }
            return undefined;
        };
    }
    /**
     * @param factory factory of blocks
     * @param v empty value of type E (only used for type inference)
     * @return new empty list.
     */
    static emptyWith(factory) {
        return new EditableOpLinkedList(new Sentinel(), 0, factory);
    }
    /** @Override */
    anchorAt(index, isAfter) {
        assert(() => isU32(index), "index ∈ u32");
        return this.root.anchorAt(index, isAfter, this.factory);
    }
    /** @Override */
    remove(delta) {
        this.factory.garbageCollect(delta);
        return super.remove(delta);
    }
    // Modification
    /** @Override */
    insertAt(index, items) {
        assert(() => isU32(index), "index ∈ u32");
        assert(() => index <= this.length, "valid index");
        assert(() => items.length > 0, "items is not empty");
        const result = this.root.insertAt(index, items, this.factory);
        this.length = this.length + items.length;
        return result;
    }
    /** @Override */
    removeAt(index, length) {
        assert(() => isU32(index), "index ∈ u32");
        assert(() => isU32(length), "length ∈ u32");
        assert(() => index + length <= this.length, "valid end index");
        this.length = this.length - length;
        const rmv = this.root.removeAt(index, length);
        if (this.factory.canGarbageCollect()) {
            for (const dBlock of rmv) {
                this.factory.garbageCollect(dBlock);
            }
        }
        return rmv;
    }
}
//# sourceMappingURL=op-linked-list.js.map