"use strict";
/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var assert_1 = require("../util/assert");
var block_1 = require("./block");
var anchor_1 = require("./anchor");
/**
 * Factory of blocks.
 * Implementations can implement different strategies of generation.
 *
 * Implementations should have at least two static function described in
 * {@see BlockFactoryConstructor }.
 */
var BlockFactory = /** @class */ (function () {
    /**
     * @param posBounds bottom and top positions.
     */
    function BlockFactory(posBounds) {
        this.posBounds = posBounds;
        this.bottomAnchor = anchor_1.Anchor.from(posBounds.BOTTOM, true);
        this.topAnchor = anchor_1.Anchor.from(posBounds.TOP, false);
    }
    /**
     * @param items
     * @return New block with {@link items } as {@link Block#items }.
     */
    BlockFactory.prototype.from = function (items) {
        assert_1.assert(function () { return items.length > 0; }, "items.length > 0");
        var pos = this.posBetween(this.posBounds.BOTTOM, items.length, this.posBounds.TOP);
        return new block_1.Block(pos, items);
    };
    /**
     * @param l
     * @param items
     * @return New block after {@link l } with {@link items } as
     *  {@link Block#items }.
     */
    BlockFactory.prototype.after = function (l, items) {
        assert_1.assert(function () { return items.length > 0; }, "items.length > 0");
        var pos = this.posBetween(l.upperPos(), items.length, this.posBounds.TOP);
        return new block_1.Block(pos, items);
    };
    /**
     * @param items
     * @param u
     * @return New block before {@link u } with {@link items } as
     *  {@link Block#items }.
     */
    BlockFactory.prototype.before = function (items, u) {
        assert_1.assert(function () { return items.length > 0; }, "items.length > 0");
        var pos = this.posBetween(this.posBounds.BOTTOM, items.length, u.lowerPos);
        return new block_1.Block(pos, items);
    };
    /**
     * @param l
     * @param items
     * @param u
     * @return New block after {@link l } and before {@link u }
     *  with {@link items } as {@link Block#items }.
     */
    BlockFactory.prototype.between = function (l, items, u) {
        assert_1.assert(function () { return items.length > 0; }, "items.length > 0");
        assert_1.heavyAssert(function () {
            return l === undefined ||
                u === undefined ||
                l.compare(u) <= -6 /* PREPENDABLE */;
        }, "l < u");
        var lPos = l !== undefined ? l.upperPos() : this.posBounds.BOTTOM;
        var uPos = u !== undefined ? u.lowerPos : this.posBounds.TOP;
        var pos = this.posBetween(lPos, items.length, uPos);
        return new block_1.Block(pos, items);
    };
    /**
     * See {@link BlockFactory#garbageCollect }
     *
     * @param dBlock
     * @return Is there stored positions in this factory?
     */
    BlockFactory.prototype.canGarbageCollect = function () {
        return (this.garbageCollectPos !== BlockFactory.prototype.garbageCollectPos);
    };
    /**
     * Some factories can store generated positions for differents purposes.
     * This method enables to garbage collect these positions when they
     * are removed from the main data structure.
     *
     * @param dBlock
     */
    BlockFactory.prototype.garbageCollect = function (dBlock) {
        this.garbageCollectPos(dBlock.lowerPos, dBlock.length);
    };
    /**
     * See {@link BlockFactory#garbageCollect }
     *
     * @param l lower position
     * @param length number of positions after l (including l) to remove
     */
    BlockFactory.prototype.garbageCollectPos = function (l, length) { };
    return BlockFactory;
}());
exports.BlockFactory = BlockFactory;
//# sourceMappingURL=block-factory.js.map