"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var block_list_context_1 = require("./block-list-context");
var assert_1 = require("../util/assert");
var number_1 = require("../util/number");
var data_validation_1 = require("../util/data-validation");
var DeltaReplicatedList = /** @class */ (function () {
    function DeltaReplicatedList(list, vv) {
        this.list = list;
        this.versionVector = vv;
    }
    DeltaReplicatedList.from = function (list) {
        assert_1.assert(function () { return list.length === 0; }, "list must be empty");
        return new DeltaReplicatedList(list, Object.create(null));
    };
    /**
     * @param opFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    DeltaReplicatedList.fromPlain = function (opFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x) &&
                data_validation_1.isObject(x.versionVector)) {
                var versionVector = Object.create(null);
                for (var replica in x.versionVector) {
                    if (x.versionVector.hasOwnProperty(replica)) {
                        var seq = x.versionVector[replica];
                        if (number_1.isU32(seq)) {
                            versionVector[replica] = seq;
                        }
                    }
                }
                var list = opFromPlain(x.list);
                if (list !== undefined) {
                    return new DeltaReplicatedList(list, versionVector);
                }
            }
            return undefined;
        };
    };
    DeltaReplicatedList.prototype.readonlyVersionvector = function () {
        return this.versionVector;
    };
    Object.defineProperty(DeltaReplicatedList.prototype, "length", {
        get: function () {
            return block_list_context_1.BlockListContext.length;
        },
        enumerable: true,
        configurable: true
    });
    DeltaReplicatedList.prototype.reduceBlock = function (f, prefix) {
        return this.list.reduceBlock(f, prefix);
    };
    /**
     * @param prefix
     * @return Concatenated version prefixed by `prefix`.
     */
    DeltaReplicatedList.prototype.concatenated = function (prefix) {
        return this.list.concatenated(prefix);
    };
    /**
     * Non-cryptographic way to approximate object identity.
     * Do not take the blocks' content into account.
     */
    DeltaReplicatedList.prototype.structuralHashCode = function () {
        return this.list.structuralHashCode();
    };
    DeltaReplicatedList.prototype.lastSeq = function (replica) {
        var maybeLastSeq = this.versionVector[replica];
        if (maybeLastSeq !== undefined) {
            return maybeLastSeq;
        }
        return 0;
    };
    /**
     * @param anchor
     * @return index of `anchor`.
     */
    DeltaReplicatedList.prototype.indexFrom = function (anchor) {
        return this.list.indexFrom(anchor);
    };
    /**
     * Complexity: O(1)
     *
     * @param delta
     * @return parts of `delta` that can be inserted.
     */
    DeltaReplicatedList.prototype.insertable = function (delta) {
        var lastSeq = this.lastSeq(String(delta.replica()));
        var deltaSeqs = delta.seqs();
        if (deltaSeqs.lower > lastSeq) {
            return [delta];
        }
        else if (deltaSeqs.upper() > lastSeq) {
            var remaining = delta.rightSplitAt(lastSeq + 1 - deltaSeqs.lower);
            return [remaining];
        }
        return [];
    };
    /**
     * Complexity: O(S) where S depends of the chosen implementation.
     * e.g. for a balanced tree: S = log2(N) with N the number of blocks.
     *
     * @param delta
     * @return parts of `delta` that was removed.
     */
    DeltaReplicatedList.prototype.removed = function (delta) {
        var lastSeq = this.lastSeq(String(delta.replica()));
        var deltaSeqs = delta.seqs();
        if (deltaSeqs.lower > lastSeq) {
            return [];
        }
        else if (deltaSeqs.upper() > lastSeq) {
            var splittingIndex = lastSeq + 1 - deltaSeqs.lower;
            var remaining = delta.leftSplitAt(splittingIndex);
            return this.list.insertable(remaining).map(function (block) {
                return block.toLengthBlock();
            });
        }
        return this.list.insertable(delta).map(function (block) {
            return block.toLengthBlock();
        });
    };
    DeltaReplicatedList.prototype.bumpVersion = function (delta) {
        var lastSeq = this.lastSeq(String(delta.replica()));
        var deltaSeqs = delta.seqs();
        if (deltaSeqs.lower > lastSeq + 1) {
            throw new Error("FIFO violation");
        }
        else {
            this.versionVector[delta.replica()] = Math.max(lastSeq, deltaSeqs.upper());
        }
    };
    DeltaReplicatedList.prototype.insert = function (delta) {
        var e_1, _a;
        var result = [];
        try {
            for (var _b = tslib_1.__values(this.insertable(delta)), _c = _b.next(); !_c.done; _c = _b.next()) {
                var d = _c.value;
                var ins = this.list.insert(d);
                result = result.concat(ins);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.bumpVersion(delta);
        return result;
    };
    DeltaReplicatedList.prototype.remove = function (delta) {
        this.bumpVersion(delta);
        return this.list.remove(delta);
    };
    DeltaReplicatedList.prototype.quickFilter = function (delta) {
        if (delta.isLengthBlock()) {
            return [delta];
        }
        else {
            return this.insertable(delta);
        }
    };
    DeltaReplicatedList.prototype.applyDelta = function (delta) {
        if (delta.isLengthBlock()) {
            return this.remove(delta);
        }
        else {
            return this.insert(delta);
        }
    };
    /**
     * [Mutation]
     * Unidirectional merge from other to this.
     * This operation is idempotent.
     *
     * @param other
     * @return Performed modifications in terms of local operations.
     *  The n+1 -th operation depends on the n -th operation.
     */
    DeltaReplicatedList.prototype.merge = function (other) {
        var e_2, _a, e_3, _b;
        var _this = this;
        // determine inserted blocks in `other` and missing in `this`
        var inserted = other.reduceBlock(function (acc, block) {
            return acc.concat(_this.insertable(block));
        }, []);
        // determine removed block in `other` and present in `this`
        var removed = this.reduceBlock(function (acc, block) {
            return acc.concat(other.removed(block));
        }, []);
        // apply raw insertions and raw removals (bypassing causal context)
        var result = [];
        try {
            for (var inserted_1 = tslib_1.__values(inserted), inserted_1_1 = inserted_1.next(); !inserted_1_1.done; inserted_1_1 = inserted_1.next()) {
                var iBlock = inserted_1_1.value;
                result = result.concat(this.list.insert(iBlock));
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (inserted_1_1 && !inserted_1_1.done && (_a = inserted_1.return)) _a.call(inserted_1);
            }
            finally { if (e_2) throw e_2.error; }
        }
        try {
            for (var removed_1 = tslib_1.__values(removed), removed_1_1 = removed_1.next(); !removed_1_1.done; removed_1_1 = removed_1.next()) {
                var dBlock = removed_1_1.value;
                result = result.concat(this.list.remove(dBlock));
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (removed_1_1 && !removed_1_1.done && (_b = removed_1.return)) _b.call(removed_1);
            }
            finally { if (e_3) throw e_3.error; }
        }
        // merge causal contexts
        for (var replica in other.readonlyVersionvector()) {
            var otherLastSeq = other.readonlyVersionvector()[replica];
            var lastSeq = this.versionVector[replica];
            if (lastSeq !== undefined && otherLastSeq !== undefined) {
                this.versionVector[replica] = Math.max(lastSeq, otherLastSeq);
            }
            else if (otherLastSeq !== undefined) {
                this.versionVector[replica] = otherLastSeq;
            }
        }
        return result;
    };
    return DeltaReplicatedList;
}());
exports.DeltaReplicatedList = DeltaReplicatedList;
var DeltaEditableReplicatedList = /** @class */ (function (_super) {
    tslib_1.__extends(DeltaEditableReplicatedList, _super);
    function DeltaEditableReplicatedList(list, vv) {
        var _this = _super.call(this, list, vv) || this;
        _this.list = list;
        return _this;
    }
    DeltaEditableReplicatedList.from = function (list) {
        assert_1.assert(function () { return list.length === 0; }, "list must be empty");
        return new DeltaEditableReplicatedList(list, Object.create(null));
    };
    /**
     * @param opFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    DeltaEditableReplicatedList.fromPlain = function (opFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x) &&
                data_validation_1.isObject(x.versionVector)) {
                var versionVector = Object.create(null);
                for (var replica in x.versionVector) {
                    if (x.versionVector.hasOwnProperty(replica)) {
                        var seq = x.versionVector[replica];
                        if (number_1.isU32(seq)) {
                            versionVector[replica] = seq;
                        }
                    }
                }
                var list = opFromPlain(x.list);
                if (list !== undefined) {
                    return new DeltaEditableReplicatedList(list, versionVector);
                }
            }
            return undefined;
        };
    };
    /**
     * @param index index where the anchor is
     * @param isAfter Is the anchor after `index`?
     * @return anchor at `index`.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    DeltaEditableReplicatedList.prototype.anchorAt = function (index, isAfter) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        return this.list.anchorAt(index, isAfter);
    };
    /**
     * [Mutation]
     * Insert {@link items } at {@link index}.
     *
     * @param index 0-based index; Where to insert.
     * @param items elements to insert.
     * @return Delta which represents the insertion.
     */
    DeltaEditableReplicatedList.prototype.insertAt = function (index, items) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        assert_1.assert(function () { return items.length > 0; }, "items is not empty");
        var iBlock = this.list.insertAt(index, items);
        this.versionVector[iBlock.replica()] = iBlock.seqs().upper();
        return iBlock;
    };
    /**
     * [Mutation]
     * Remove a number of {@link length } elements from {@link index }
     *
     * @param index 0-based index.
     * @param length Number of elements to remove.
     * @return Delta which represents the deletion.
     */
    DeltaEditableReplicatedList.prototype.removeAt = function (index, length) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        assert_1.assert(function () { return number_1.isU32(length); }, "length ∈ u32");
        return this.list.removeAt(index, length);
    };
    return DeltaEditableReplicatedList;
}(DeltaReplicatedList));
exports.DeltaEditableReplicatedList = DeltaEditableReplicatedList;
//# sourceMappingURL=delta-replicated-list.js.map