"use strict";
/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var assert_1 = require("../../util/assert");
var data_validation_1 = require("../../util/data-validation");
var number_1 = require("../../util/number");
var ordering_1 = require("../../util/ordering");
/**
 * @final this class cannot be extended
 *
 * Part of a {@link SimpleDotPos}.
 *
 * A part is a lexicographic triplet (priority, replica, seq).
 * The dot (replica, seq) uniquely identifies.
 */
var SimpleDotPosPart = /** @class */ (function () {
    /**
     * @param priority {@link SimpleDotPosPart#priority }
     * @param replica {@link SimpleDotPosPart#replica }
     * @param seq {@link SimpleDotPosPart#seq }
     */
    function SimpleDotPosPart(priority, replica, seq) {
        assert_1.assert(function () { return number_1.isU32(priority); }, "random ∈ u32");
        assert_1.assert(function () { return number_1.isU32(replica); }, "replica ∈ u32");
        assert_1.assert(function () { return number_1.isU32(seq); }, "seq ∈ u32");
        this.priority = priority;
        this.replica = replica;
        this.seq = seq;
    }
    /**
     * @param priority {@link SimpleDotPosPart#priority }
     *      ∉ {U32_BOTTOM, U32_TOP}
     * @param replica {@link SimpleDotPosPart#replica }
     *      !== U32_TOP
     * @param seq {@link SimpleDotPosPart#seq }
     */
    SimpleDotPosPart.from = function (priority, replica, seq) {
        assert_1.assert(function () { return priority !== number_1.U32_BOTTOM && priority !== number_1.U32_TOP; }, "priority ∉ {U32_BOTTOM, U32_TOP}");
        assert_1.assert(function () { return replica !== number_1.U32_TOP; }, "replica != U32_TOP. Reserved for BOTTOM and TOP.");
        assert_1.assert(function () { return seq !== 0; }, "seq != 0. Reserved for convenience.");
        return new SimpleDotPosPart(priority, replica, seq);
    };
    /**
     * @note {@link SimpleDotPosPart#BOTTOM} and {@link SimpleDotPosPart#TOP}
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    SimpleDotPosPart.fromPlain = function (x) {
        if (data_validation_1.isObject(x) &&
            number_1.isU32(x.priority) &&
            number_1.isU32(x.replica) &&
            number_1.isU32(x.seq) &&
            x.seq !== 0) {
            return new SimpleDotPosPart(x.priority, x.replica, x.seq);
        }
        return undefined;
    };
    // Derivation
    /**
     * @param seq The seq of the new SimpleDotPosPart
     * @return part with the same base,
     *  but with a different seq
     */
    SimpleDotPosPart.prototype.withSeq = function (seq) {
        assert_1.assert(function () { return number_1.isU32(seq); }, "seq ∈ u32");
        return new SimpleDotPosPart(this.priority, this.replica, seq);
    };
    /**
     * @return [priority, replica, seq]
     */
    SimpleDotPosPart.prototype.asTuple = function () {
        return [this.priority, this.replica, this.seq];
    };
    /**
     * Non-cryptographic way to approximate object identity.
     */
    SimpleDotPosPart.prototype.hashCode = function () {
        return number_1.hashCodeOf(this.asTuple());
    };
    // Status
    /**
     * @example
     * SimpleDotPosPart(0, 0, _) == SimpleDotPosPart(0, 0, _)
     * SimpleDotPosPart(0, 0, _) < SimpleDotPosPart(1, 0, _)
     * SimpleDotPosPart(1, 0, _) < SimpleDotPosPart(1, 1, _)
     *
     * Lexicographic order of pair (priority, replica).
     *
     * @param other
     * @return this [Order relation] {@link other }, regardless their seqs
     *  ({@link SimpleDotPosPart#seq}).
     */
    SimpleDotPosPart.prototype.compareBase = function (other) {
        return ordering_1.lexCompareOrdering(number_1.compareU32(this.priority, other.priority), number_1.compareU32(this.replica, other.replica));
    };
    /**
     * @example
     * SimpleDotPosPart(0, 0, 0) < SimpleDotPosPart(1, 0, 0)
     * SimpleDotPosPart(1, 0, 0) < SimpleDotPosPart(1, 1, 0)
     * SimpleDotPosPart(1, 1, 0) < SimpleDotPosPart(1, 1, 1)
     *
     * Lexicographic order of triple (priority, replica, seq).
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    SimpleDotPosPart.prototype.compare = function (other) {
        if (this === other) {
            return 0 /* EQUAL */;
        }
        else {
            return ordering_1.lexCompareOrdering(this.compareBase(other), number_1.compareU32(this.seq, other.seq));
        }
    };
    /**
     * Lowest part.
     * Here the pair (replica, seq) has no meaning.
     */
    SimpleDotPosPart.BOTTOM = new SimpleDotPosPart(number_1.U32_BOTTOM, number_1.U32_TOP, 1);
    /**
     * Greatest part.
     * Here the pair (replica, seq) has no meaning.
     */
    SimpleDotPosPart.TOP = new SimpleDotPosPart(number_1.U32_TOP, number_1.U32_TOP, 2);
    return SimpleDotPosPart;
}());
exports.SimpleDotPosPart = SimpleDotPosPart;
//# sourceMappingURL=simple-dot-pos-part.js.map