"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var local_operation_1 = require("../../core/local-operation");
var assert_1 = require("../../util/assert");
var block_list_context_1 = require("../../core/block-list-context");
var data_validation_1 = require("../../util/data-validation");
function lengthOf(node) {
    if (node !== undefined) {
        return node.length;
    }
    return 0;
}
exports.lengthOf = lengthOf;
function rankOf(node) {
    if (node !== undefined) {
        return node.rank;
    }
    return 0;
}
var ValuedNode = /** @class */ (function (_super) {
    tslib_1.__extends(ValuedNode, _super);
    function ValuedNode(l, block, r) {
        var _this = _super.call(this) || this;
        _this.left = l;
        _this.block = block;
        _this.right = r;
        _this.isSelfRemoved = false;
        _this.length = 0;
        _this.rank = 0;
        _this.update();
        return _this;
    }
    ValuedNode.leaf = function (block) {
        return new ValuedNode(undefined, block, undefined);
    };
    /**
     * @param blockFromPlain
     * @return function that accepts a value and returns a Node from
     *  the value, or undefined if the value is mal-formed
     */
    ValuedNode.fromPlain = function (blockFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x)) {
                var left = ValuedNode.fromPlain(blockFromPlain)(x.left);
                var right = ValuedNode.fromPlain(blockFromPlain)(x.right);
                var block = blockFromPlain(x.block);
                if (block !== undefined) {
                    return new ValuedNode(left, block, right);
                }
            }
            return undefined;
        };
    };
    // Basic changes
    ValuedNode.prototype.setLeft = function (l) {
        ;
        this.left = l;
        this.update();
    };
    ValuedNode.prototype.replace = function (block) {
        ;
        this.block = block;
        this.update();
    };
    ValuedNode.prototype.setRight = function (r) {
        ;
        this.right = r;
        this.update();
    };
    ValuedNode.prototype.update = function () {
        this.length =
            lengthOf(this.left) + this.block.length + lengthOf(this.right);
        this.rank = 1 + Math.max(rankOf(this.left), rankOf(this.right));
    };
    // Balancing
    ValuedNode.prototype.balanceFactor = function () {
        return rankOf(this.right) - rankOf(this.left);
    };
    ValuedNode.prototype.isBalanced = function () {
        var balanceFactor = this.balanceFactor();
        return -2 < balanceFactor && balanceFactor < 2;
    };
    ValuedNode.prototype.isRightUnbalanced = function () {
        return this.balanceFactor() >= 2;
    };
    ValuedNode.prototype.isRightOriented = function () {
        return this.balanceFactor() === 1;
    };
    ValuedNode.prototype.isLeftOriented = function () {
        return this.balanceFactor() <= -2;
    };
    ValuedNode.prototype.isSlighlyLeftUnbalanced = function () {
        return this.balanceFactor() === -1;
    };
    ValuedNode.prototype.rotateLeft = function () {
        var r = this.right;
        this.setRight(r.left);
        r.setLeft(this);
        return r;
    };
    ValuedNode.prototype.rotateRight = function () {
        var l = this.left;
        this.setLeft(l.right);
        l.setRight(this);
        return l;
    };
    ValuedNode.prototype.balance = function () {
        var self = this;
        while (self.isRightUnbalanced()) {
            if (self.right.isSlighlyLeftUnbalanced()) {
                self.setRight(self.right.rotateRight());
            }
            self = self.rotateLeft();
        }
        while (self.isLeftOriented()) {
            if (self.left.isRightOriented()) {
                self.setLeft(self.left.rotateLeft());
            }
            self = self.rotateRight();
        }
        return self;
    };
    // Self removal
    ValuedNode.prototype.updated = function () {
        if (this.isSelfRemoved) {
            return undefined;
        }
        return this.balance();
    };
    // Traversal
    ValuedNode.prototype.reduceBlock = function (f, prefix) {
        var leftPrefix = prefix;
        if (this.left !== undefined) {
            leftPrefix = this.left.reduceBlock(f, prefix);
        }
        leftPrefix = f(leftPrefix, this.block);
        if (this.right !== undefined) {
            return this.right.reduceBlock(f, leftPrefix);
        }
        return leftPrefix;
    };
    // Impl
    ValuedNode.prototype.current = function () {
        return this.block;
    };
    ValuedNode.prototype.currentRelativeIndex = function () {
        return lengthOf(this.left);
    };
    ValuedNode.prototype.min = function () {
        if (this.left !== undefined) {
            return this.left.min(); // tail recursion
        }
        return this.block;
    };
    ValuedNode.prototype.max = function () {
        if (this.right !== undefined) {
            return this.right.max(); // tail recursion
        }
        return this.block;
    };
    ValuedNode.prototype.predecessor = function () {
        if (this.left !== undefined) {
            return this.left.max();
        }
        return undefined;
    };
    ValuedNode.prototype.successor = function () {
        if (this.right !== undefined) {
            return this.right.min();
        }
        return undefined;
    };
    ValuedNode.prototype.insertMin = function (iBlock) {
        var _this = this;
        if (this.left !== undefined) {
            this.setLeft(this.left.insertMin(iBlock));
        }
        else {
            assert_1.heavyAssert(function () { return _this.block.compare(iBlock) === 7 /* AFTER */; }, "iBlock < block");
            this.setLeft(ValuedNode.leaf(iBlock));
        }
        return this.balance();
    };
    ValuedNode.prototype.insertMax = function (iBlock) {
        var _this = this;
        if (this.right !== undefined) {
            this.setRight(this.right.insertMax(iBlock));
        }
        else {
            assert_1.heavyAssert(function () { return _this.block.compare(iBlock) === -7 /* BEFORE */; }, "block < iBlock");
            this.setRight(ValuedNode.leaf(iBlock));
        }
        return this.balance();
    };
    // Anchor
    /** @Override */
    ValuedNode.prototype.indexFromLeft = function (anchor, minIndex) {
        if (this.left !== undefined) {
            return this.left.indexFrom(anchor, minIndex);
        }
        return minIndex;
    };
    /** @Override */
    ValuedNode.prototype.indexFromRight = function (anchor, minIndex) {
        if (this.right !== undefined) {
            return this.right.indexFrom(anchor, minIndex);
        }
        return minIndex;
    };
    /** @Override */
    ValuedNode.prototype.anchorAtLeft = function (relIndex, isAfter, f) {
        if (this.left !== undefined) {
            return this.left.anchorAt(relIndex, isAfter, f);
        }
        return f.bottomAnchor;
    };
    /** @Override */
    ValuedNode.prototype.anchorAtRight = function (relIndex, isAfter, f) {
        if (this.right !== undefined) {
            return this.right.anchorAt(relIndex, isAfter, f);
        }
        else if (isAfter) {
            return f.topAnchor;
        }
        else {
            return this.block.upperAnchor();
        }
    };
    // Insertion
    /** @Override */
    ValuedNode.prototype.insertableLeft = function (iBlock) {
        if (this.left !== undefined) {
            return this.left.insertable(iBlock);
        }
        return [iBlock];
    };
    /** @Override */
    ValuedNode.prototype.insertableRight = function (iBlock) {
        if (this.right !== undefined) {
            return this.right.insertable(iBlock);
        }
        return [iBlock];
    };
    /** @Override */
    ValuedNode.prototype.insertPredecessor = function (iBlock) {
        if (this.left !== undefined) {
            this.setLeft(this.left.insertMax(iBlock));
        }
        else {
            this.setLeft(ValuedNode.leaf(iBlock));
        }
        // delegate balance to parent
    };
    /** @Override */
    ValuedNode.prototype.insertSuccessor = function (iBlock) {
        if (this.right !== undefined) {
            this.setRight(this.right.insertMin(iBlock));
        }
        else {
            this.setRight(ValuedNode.leaf(iBlock));
        }
        // delegate balance to parent
    };
    /** @Override */
    ValuedNode.prototype.insertLeft = function (iBlock, minIndex) {
        var ins;
        if (this.left !== undefined) {
            ins = this.left.insert(iBlock, minIndex);
            this.setLeft(this.left.balance());
        }
        else {
            this.setLeft(ValuedNode.leaf(iBlock));
            ins = [new local_operation_1.Ins(minIndex, iBlock.content)];
        }
        // delegate balance to parent
        return ins;
    };
    /** @Override */
    ValuedNode.prototype.insertRight = function (iBlock, minIndex) {
        var ins;
        if (this.right !== undefined) {
            ins = this.right.insert(iBlock, minIndex);
            this.setRight(this.right.balance());
        }
        else {
            // should not happen
            this.setRight(ValuedNode.leaf(iBlock));
            ins = [new local_operation_1.Ins(minIndex, iBlock.content)];
        }
        // delegate balance to parent
        return ins;
    };
    /** @Override */
    ValuedNode.prototype.insertAtLeft = function (idx, items, f) {
        var ins;
        if (this.left !== undefined) {
            ins = this.left.insertAt(idx, items, f);
            this.setLeft(this.left.balance());
        }
        else {
            // should not happen
            ins = this.insertAt(0, items, f);
        }
        // delegate balance to parent
        return ins;
    };
    /** @Override */
    ValuedNode.prototype.insertAtRight = function (idx, items, f) {
        var ins;
        if (this.right !== undefined) {
            ins = this.right.insertAt(idx, items, f);
            this.setRight(this.right.balance());
        }
        else {
            var lastIdx = this.currentRelativeIndex() + this.block.length;
            ins = this.insertAt(lastIdx, items, f);
        }
        // delegate balance to parent
        return ins;
    };
    // Removal
    /** @Override */
    ValuedNode.prototype.removeCurrent = function () {
        this.isSelfRemoved = true;
    };
    ValuedNode.prototype.removeMin = function () {
        if (this.left !== undefined) {
            this.setLeft(this.left.removeMin());
            return this.balance();
        }
        return this.right;
    };
    ValuedNode.prototype.removeMax = function () {
        if (this.right !== undefined) {
            this.setRight(this.right.removeMax());
            return this.balance();
        }
        return this.left;
    };
    /** @Override */
    ValuedNode.prototype.removePredecessor = function () {
        if (this.left !== undefined) {
            this.setLeft(this.left.removeMax());
        }
        // delegate balance to parent
    };
    /** @Override */
    ValuedNode.prototype.removeSuccessor = function () {
        if (this.right !== undefined) {
            this.setRight(this.right.removeMin());
        }
        // delegate balance to parent
    };
    /** @Override */
    ValuedNode.prototype.removeLeft = function (dBlock, minIndex) {
        if (this.left !== undefined) {
            var rmv = this.left.remove(dBlock, minIndex);
            this.setLeft(this.left.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    };
    /** @Override */
    ValuedNode.prototype.removeRight = function (dBlock, minIndex) {
        if (this.right !== undefined) {
            var rmv = this.right.remove(dBlock, minIndex);
            this.setRight(this.right.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    };
    /** @Override */
    ValuedNode.prototype.removeAtLeft = function (dRange, minIndex) {
        if (this.left !== undefined) {
            var rmv = this.left.removeAt(dRange, minIndex);
            this.setLeft(this.left.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    };
    /** @Override */
    ValuedNode.prototype.removeAtRight = function (dRange, minIndex) {
        if (this.right !== undefined) {
            var rmv = this.right.removeAt(dRange, minIndex);
            this.setRight(this.right.updated());
            // delegate balance to parent
            return rmv;
        }
        return [];
    };
    return ValuedNode;
}(block_list_context_1.BlockListContext));
exports.ValuedNode = ValuedNode;
//# sourceMappingURL=avl-list-node.js.map