"use strict";
/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var assert_1 = require("../../util/assert");
var local_operation_1 = require("../../core/local-operation");
var number_1 = require("../../util/number");
var data_validation_1 = require("../../util/data-validation");
var Linkable = /** @class */ (function () {
    /**
     * @param right Next cell.
     */
    function Linkable(right) {
        this.right = right;
    }
    // Access
    /**
     * @param f reducer that respectively accepts the accumulated value and
     *  the current block as first and second parameters. It returns the next
     *  accumulated value.
     * @param prefix initial value
     * @return accumulated value by reducing blocks from left to right
     *  in the list.
     */
    Linkable.prototype.reduceBlock = function (f, prefix) {
        if (this.right === undefined) {
            return prefix;
        }
        else {
            var rBlock = this.right.block;
            return this.right.reduceBlock(f, f(prefix, rBlock)); // Tail recursion
        }
    };
    /**
     * @param anchor
     * @param minIndex minimal index in the current sub-list.
     * @return index of `anchor`.
     */
    Linkable.prototype.indexFrom = function (anchor, minIndex) {
        assert_1.assert(function () { return number_1.isU32(minIndex); }, "minIndex ∈ u32");
        if (this.right !== undefined) {
            var rBlock = this.right.block;
            var rBlockIndex = rBlock.indexFrom(anchor);
            if (rBlockIndex === rBlock.length) {
                return this.right.indexFrom(anchor, minIndex + rBlock.length);
            }
            return minIndex + rBlockIndex;
        }
        return minIndex;
    };
    /**
     * @param relIndex relative index where the anchor is
     * @param isAfter Is the anchor after or before `relIndex`?
     * @param f factory
     * @return anchor at `index` in the current list.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    Linkable.prototype.anchorAt = function (index, isAfter, f) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        if (this.right !== undefined) {
            var rBlock = this.right.block;
            if (index >= rBlock.length) {
                return this.right.anchorAt(index - rBlock.length, isAfter, f);
            }
            else if (index !== 0 || isAfter) {
                return rBlock.anchor(index, !isAfter);
            }
        }
        else if (isAfter) {
            return f.topAnchor;
        }
        if (this instanceof Cell) {
            return this.block.upperAnchor();
        }
        else {
            return f.bottomAnchor;
        }
    };
    /**
     * @param iBlock
     * @return segments of `iBlock` that can be inserted in the current
     *  sub-list.
     */
    Linkable.prototype.insertable = function (iBlock) {
        if (this.right === undefined) {
            return [iBlock];
        }
        else {
            var rBlock = this.right.block;
            switch (iBlock.compare(rBlock)) {
                case -6 /* PREPENDABLE */:
                case -7 /* BEFORE */:
                case 5 /* SPLITTING */:
                    return [iBlock];
                case 6 /* APPENDABLE */:
                case 7 /* AFTER */:
                    return this.right.insertable(iBlock);
                case -5 /* SPLITTED_BY */: {
                    var _a = tslib_1.__read(iBlock.splitWith(rBlock), 2), lSplit = _a[0], rSplit = _a[1];
                    var rIns = this.right.insertable(rSplit);
                    return [lSplit].concat(rIns);
                }
                case 3 /* INCLUDING_RIGHT */:
                case -4 /* OVERLAPPING_BEFORE */:
                    return [iBlock.prependable(rBlock)];
                case 1 /* INCLUDING_LEFT */:
                case 4 /* OVERLAPPING_AFTER */:
                    return this.right.insertable(iBlock.appendable(rBlock));
                case 2 /* INCLUDING_MIDDLE */: {
                    var lIns = [iBlock.prependable(rBlock)];
                    var rIns = this.right.insertable(iBlock.appendable(rBlock));
                    return lIns.concat(rIns);
                }
                case -2 /* INCLUDED_MIDDLE_BY */:
                case -1 /* INCLUDED_RIGHT_BY */:
                case -3 /* INCLUDED_LEFT_BY */:
                case 0 /* EQUAL */:
                    return [];
                default:
                    throw new Error("non-exhaustive switch");
            }
        }
    };
    // Modification
    /**
     * [Mutation]
     * Insert {@link rblock } to right and preserve the chain.
     *
     * @param rblock block to insert.
     * @return inserted cell with {@link rblock } as content.
     */
    Linkable.prototype.insertRight = function (rblock) {
        this.right = new Cell(rblock, this.right);
        return this.right;
    };
    /**
     * [Mutation]
     * Insert the parts of {@link iBlock } which are not already inserted.
     *
     * @param iBlock block to insert.
     * @param index 0-based index
     * @return Performed modifications in terms of local insertions.
     *  The n+1 -th insertion depends on the effect of the n -th insertion.
     */
    Linkable.prototype.insert = function (iBlock, index) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        if (this.right === undefined) {
            this.insertRight(iBlock);
            return [new local_operation_1.Ins(index, iBlock.content)];
        }
        else {
            var rBlock = this.right.block;
            switch (rBlock.compare(iBlock)) {
                case -7 /* BEFORE */:
                    return this.right.insert(iBlock, index + rBlock.length);
                case -6 /* PREPENDABLE */: {
                    var insertions = this.right.insert(iBlock, index + rBlock.length);
                    // The right block of rBlock could be a splitting block
                    this.right = this.right.right;
                    this.insert(rBlock, 0); // handle prepend
                    return insertions;
                }
                case 7 /* AFTER */:
                    this.insertRight(iBlock);
                    return [new local_operation_1.Ins(index, iBlock.content)];
                case 6 /* APPENDABLE */:
                    this.right = this.right.right;
                    this.insertRight(iBlock.append(rBlock));
                    return [new local_operation_1.Ins(index, iBlock.content)];
                case 5 /* SPLITTING */: {
                    var _a = tslib_1.__read(iBlock.splitWith(rBlock), 2), lSplit = _a[0], rSplit = _a[1];
                    var rInsertions = this.right.insert(rSplit, index + lSplit.length + rBlock.length);
                    this.insertRight(lSplit);
                    return tslib_1.__spread([new local_operation_1.Ins(index, lSplit.content)], rInsertions);
                }
                case -5 /* SPLITTED_BY */: {
                    var _b = tslib_1.__read(rBlock.splitWith(iBlock), 2), lSplit = _b[0], rSplit = _b[1];
                    this.right = this.right.right;
                    this.insertRight(lSplit)
                        .insertRight(iBlock)
                        .insertRight(rSplit);
                    return [new local_operation_1.Ins(index + lSplit.length, iBlock.content)];
                }
                case -4 /* OVERLAPPING_BEFORE */:
                case 4 /* OVERLAPPING_AFTER */:
                case -2 /* INCLUDED_MIDDLE_BY */:
                case -1 /* INCLUDED_RIGHT_BY */:
                case -3 /* INCLUDED_LEFT_BY */:
                case 2 /* INCLUDING_MIDDLE */:
                case 3 /* INCLUDING_RIGHT */:
                case 1 /* INCLUDING_LEFT */:
                case 0 /* EQUAL */:
                    throw new Error("inserted block intersects with one or more existing blocks");
                default:
                    throw new Error("non-exhaustive switch");
            }
        }
    };
    /**
     * [Mutation]
     * Insert {@link items } at {@link index}.
     *
     * @param index 0-based index; Where to insert.
     * @param items elements to insert.
     * @param factory strategy of block generation.
     * @return Delta which represents the insertion.
     */
    Linkable.prototype.insertAt = function (index, items, factory) {
        var _this = this;
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        assert_1.assert(function () { return items.length > 0; }, "items.length > 0");
        var iBlock;
        if (this.right === undefined) {
            if (this instanceof Cell) {
                iBlock = factory.after(this.block, items);
            }
            else {
                iBlock = factory.from(items);
            }
            this.insertRight(iBlock);
        }
        else {
            var rBlock = this.right.block;
            if (index > rBlock.length) {
                iBlock = this.right.insertAt(index - rBlock.length, items, factory);
            }
            else if (index === rBlock.length) {
                iBlock =
                    this.right.right !== undefined
                        ? factory.between(rBlock, items, this.right.right.block)
                        : factory.after(rBlock, items);
                this.insert(iBlock, 0); // handle append and prepend
            }
            else if (index === 0) {
                assert_1.assert(function () { return _this instanceof Sentinel; }, "only Sentinel can insert at index 0");
                iBlock = factory.before(items, rBlock);
                this.insert(iBlock, 0); // handle prepend
            }
            else {
                var _a = tslib_1.__read(rBlock.splitAt(index), 2), lSplit = _a[0], rSplit = _a[1];
                iBlock = factory.between(lSplit, items, rSplit);
                this.right = this.right.right; // remove right
                this.insertRight(lSplit)
                    .insertRight(iBlock)
                    .insertRight(rSplit);
            }
        }
        return iBlock;
    };
    /**
     * [Mutation]
     * Remove a number of {@link length } elements from {@link index }
     *
     * @param index 0-based index.
     * @param length Number of elements to remove.
     * @return Delta which represents the deletion.
     */
    Linkable.prototype.removeAt = function (index, length) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        assert_1.assert(function () { return number_1.isU32(length); }, "length ∈ u32");
        assert_1.assert(function () { return length > 0; }, "length > 0");
        if (this.right !== undefined) {
            var rBlock = this.right.block;
            if (index > rBlock.length) {
                return this.right.removeAt(index - rBlock.length, length);
            }
            else if (index === rBlock.length) {
                var removeds = this.right.removeAt(0, length);
                this.right = this.right.right;
                this.insert(rBlock, 0); // handle append
                return removeds;
            }
            else if (index !== 0) {
                var _a = tslib_1.__read(rBlock.splitAt(index), 2), lSplit = _a[0], rSplit = _a[1];
                this.right = this.right.right;
                var newRight = this.insertRight(lSplit);
                newRight.insertRight(rSplit);
                return newRight.removeAt(index - lSplit.length, length);
            }
            else {
                if (rBlock.length === length) {
                    this.right = this.right.right;
                    return [rBlock.toLengthBlock()];
                }
                else if (rBlock.length < length) {
                    this.right = this.right.right;
                    var removeds = this.removeAt(index, length - rBlock.length);
                    return tslib_1.__spread([rBlock.toLengthBlock()], removeds);
                }
                else {
                    var _b = tslib_1.__read(rBlock.splitAt(length), 2), lSplit = _b[0], rSplit = _b[1];
                    this.right = this.right.right;
                    this.insertRight(rSplit);
                    return [lSplit.toLengthBlock()];
                }
            }
        }
        else {
            return [];
        }
    };
    /**
     * [Mutation]
     * Remove {@link dBlock }.
     * @param dBlock block to remove.
     * @param index index of the current block in the chain.
     * @return Performed modifications in terms of local deletions.
     *  The n+1 -th deletion depends on the n -th deletion.
     */
    Linkable.prototype.remove = function (dBlock, index) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        if (this.right !== undefined) {
            var rBlock = this.right.block;
            switch (rBlock.compare(dBlock)) {
                case 5 /* SPLITTING */: // fall-through
                case -7 /* BEFORE */: {
                    var removals = this.right.remove(dBlock, index + rBlock.length);
                    var rIndex = index + rBlock.length;
                    if (removals.length > 0 && removals[0].index === rIndex) {
                        this.right = this.right.right;
                        this.insert(rBlock, 0); // handle append
                    }
                    return removals;
                }
                case -6 /* PREPENDABLE */:
                    return this.right.remove(dBlock, index + rBlock.length); // tail recursion
                case 1 /* INCLUDING_LEFT */: {
                    var rRemaning = rBlock.appendable(dBlock);
                    this.right = this.right.right;
                    this.insertRight(rRemaning);
                    return [new local_operation_1.Del(index, dBlock.length)];
                }
                case 3 /* INCLUDING_RIGHT */: {
                    var lRemaning = rBlock.prependable(dBlock);
                    this.right = this.right.right;
                    this.insertRight(lRemaning);
                    return [new local_operation_1.Del(index + lRemaning.length, dBlock.length)];
                }
                case 2 /* INCLUDING_MIDDLE */: {
                    var rRemaning = rBlock.appendable(dBlock);
                    var lRemaning = rBlock.prependable(dBlock);
                    this.right = this.right.right;
                    this.insertRight(lRemaning).insertRight(rRemaning);
                    return [new local_operation_1.Del(index + lRemaning.length, dBlock.length)];
                }
                case 0 /* EQUAL */:
                    this.right = this.right.right;
                    return [new local_operation_1.Del(index, rBlock.length)];
                case -3 /* INCLUDED_LEFT_BY */:
                case -1 /* INCLUDED_RIGHT_BY */:
                case -2 /* INCLUDED_MIDDLE_BY */: {
                    // Append of the current block is handled in the first switch-case
                    this.right = this.right.right;
                    var rRemovals = this.remove(dBlock, index);
                    return tslib_1.__spread([new local_operation_1.Del(index, rBlock.length)], rRemovals);
                }
                case -4 /* OVERLAPPING_BEFORE */: {
                    var removed = rBlock.intersection(dBlock);
                    var remaining = rBlock.prependable(dBlock);
                    //const [remaining, ] = rBlock.remove(block) as [undefined, Block<P, E>]
                    //remaining = remaining as Block<P, E>
                    this.right = this.right.right;
                    this.insertRight(remaining);
                    this.remove(dBlock, index + remaining.length);
                    return [new local_operation_1.Del(index + remaining.length, removed.length)];
                }
                case 4 /* OVERLAPPING_AFTER */: {
                    var removed = rBlock.intersection(dBlock);
                    var remaining = rBlock.appendable(dBlock);
                    //const [, remaining] = rBlock.remove(block) as [undefined, Block<P, E>]
                    this.right = this.right.right;
                    this.insertRight(remaining);
                    return [new local_operation_1.Del(index, removed.length)];
                }
                case 7 /* AFTER */:
                case 6 /* APPENDABLE */:
                case -5 /* SPLITTED_BY */:
                    break;
                default:
                    throw new Error("non-exhaustive switch");
            }
        }
        return []; // already removed or not already inserted
    };
    return Linkable;
}());
exports.Linkable = Linkable;
var Sentinel = /** @class */ (function (_super) {
    tslib_1.__extends(Sentinel, _super);
    function Sentinel() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    /**
     * @param blockFromPlain
     * @return function that accepts a value and returns a Sentinel from
     *  the value, or undefined if the value is mal-formed
     */
    Sentinel.fromPlain = function (blockFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x)) {
                var right = void 0;
                if (x.right !== undefined) {
                    right = Cell.fromPlain(blockFromPlain)(x.right);
                    if (right === undefined) {
                        return undefined;
                    }
                }
                return new Sentinel(right);
            }
            return undefined;
        };
    };
    return Sentinel;
}(Linkable));
exports.Sentinel = Sentinel;
var Cell = /** @class */ (function (_super) {
    tslib_1.__extends(Cell, _super);
    /**
     * @param block {@link Cell#block }
     * @param right {@link Linkable#right }
     */
    function Cell(block, right) {
        var _this = this;
        assert_1.heavyAssert(function () {
            return right === undefined ||
                block.compare(right.block) === -7 /* BEFORE */;
        }, "block < right.block");
        _this = _super.call(this, right) || this;
        _this.block = block;
        return _this;
    }
    /**
     * @param blockFromPlain
     * @return function that accepts a value and returns a Cell from
     *  the value, or undefined if the value is mal-formed
     */
    Cell.fromPlain = function (blockFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x)) {
                var block = blockFromPlain(x.block);
                var right = void 0;
                if (x.right !== undefined) {
                    right = Cell.fromPlain(blockFromPlain)(x.right);
                    if (right === undefined) {
                        return undefined;
                    }
                }
                if (block !== undefined) {
                    return new Cell(block, right);
                }
            }
            return undefined;
        };
    };
    return Cell;
}(Linkable));
exports.Cell = Cell;
//# sourceMappingURL=linked-list-cell.js.map