"use strict";
/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var assert_1 = require("../../util/assert");
var number_1 = require("../../util/number");
var linked_list_cell_1 = require("./linked-list-cell");
var op_replicated_list_1 = require("../../core/op-replicated-list");
var data_validation_1 = require("../../util/data-validation");
/**
 * An {@see OpReplicatedList } that uses an AVL tree.
 */
var OpLinkedList = /** @class */ (function (_super) {
    tslib_1.__extends(OpLinkedList, _super);
    /**
     * New empty list.
     */
    function OpLinkedList(root, length) {
        var _this = _super.call(this) || this;
        _this.root = root;
        _this.length = length;
        return _this;
    }
    /**
     * @param v empty value of type E
     * @return new empty list.
     */
    OpLinkedList.empty = function () {
        return new OpLinkedList(new linked_list_cell_1.Sentinel(), 0);
    };
    /**
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    OpLinkedList.fromPlain = function (f, itemsFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x) &&
                number_1.isU32(x.length)) {
                var blockFromPlain = f.blockFromPlain(itemsFromPlain);
                var root = linked_list_cell_1.Sentinel.fromPlain(blockFromPlain)(x.root);
                if (root !== undefined) {
                    return new OpLinkedList(root, x.length);
                }
            }
            return undefined;
        };
    };
    /** @Override */
    OpLinkedList.prototype.reduceBlock = function (f, prefix) {
        if (this.root !== undefined) {
            return this.root.reduceBlock(f, prefix);
        }
        return prefix;
    };
    /** @Override */
    OpLinkedList.prototype.indexFrom = function (anchor) {
        return this.root.indexFrom(anchor, 0);
    };
    /** @Override */
    OpLinkedList.prototype.insertable = function (delta) {
        return this.root.insertable(delta);
    };
    // Modification
    /** @Override */
    OpLinkedList.prototype.insert = function (block) {
        var result = this.root.insert(block, 0);
        this.length = result.reduce(function (acc, v) { return acc + v.length; }, this.length);
        return result;
    };
    /** @Override */
    OpLinkedList.prototype.remove = function (block) {
        var result = this.root.remove(block, 0);
        this.length = result.reduce(function (acc, v) { return acc - v.length; }, this.length);
        return result;
    };
    return OpLinkedList;
}(op_replicated_list_1.OpReplicatedList));
exports.OpLinkedList = OpLinkedList;
/**
 * An {@see EditableOpReplicatedList } that uses an AVL tree.
 */
var EditableOpLinkedList = /** @class */ (function (_super) {
    tslib_1.__extends(EditableOpLinkedList, _super);
    /**
     * @param factory strategy of block generation.
     */
    function EditableOpLinkedList(root, length, factory) {
        var _this = _super.call(this, root, length) || this;
        _this.factory = factory;
        return _this;
    }
    /**
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    EditableOpLinkedList.fromPlain = function (f, itemsFromPlain) {
        return function (x) {
            if (data_validation_1.isObject(x) &&
                number_1.isU32(x.length)) {
                var blockFromPlain = f.blockFromPlain(itemsFromPlain);
                var factory = f.fromPlain(x.factory);
                var root = linked_list_cell_1.Sentinel.fromPlain(blockFromPlain)(x.root);
                if (root !== undefined && factory !== undefined) {
                    return new EditableOpLinkedList(root, x.length, factory);
                }
            }
            return undefined;
        };
    };
    /**
     * @param factory factory of blocks
     * @param v empty value of type E (only used for type inference)
     * @return new empty list.
     */
    EditableOpLinkedList.emptyWith = function (factory) {
        return new EditableOpLinkedList(new linked_list_cell_1.Sentinel(), 0, factory);
    };
    /** @Override */
    EditableOpLinkedList.prototype.anchorAt = function (index, isAfter) {
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        return this.root.anchorAt(index, isAfter, this.factory);
    };
    /** @Override */
    EditableOpLinkedList.prototype.remove = function (delta) {
        this.factory.garbageCollect(delta);
        return _super.prototype.remove.call(this, delta);
    };
    // Modification
    /** @Override */
    EditableOpLinkedList.prototype.insertAt = function (index, items) {
        var _this = this;
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        assert_1.assert(function () { return index <= _this.length; }, "valid index");
        assert_1.assert(function () { return items.length > 0; }, "items is not empty");
        var result = this.root.insertAt(index, items, this.factory);
        this.length = this.length + items.length;
        return result;
    };
    /** @Override */
    EditableOpLinkedList.prototype.removeAt = function (index, length) {
        var e_1, _a;
        var _this = this;
        assert_1.assert(function () { return number_1.isU32(index); }, "index ∈ u32");
        assert_1.assert(function () { return number_1.isU32(length); }, "length ∈ u32");
        assert_1.assert(function () { return index + length <= _this.length; }, "valid end index");
        this.length = this.length - length;
        var rmv = this.root.removeAt(index, length);
        if (this.factory.canGarbageCollect()) {
            try {
                for (var rmv_1 = tslib_1.__values(rmv), rmv_1_1 = rmv_1.next(); !rmv_1_1.done; rmv_1_1 = rmv_1.next()) {
                    var dBlock = rmv_1_1.value;
                    this.factory.garbageCollect(dBlock);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (rmv_1_1 && !rmv_1_1.done && (_a = rmv_1.return)) _a.call(rmv_1);
                }
                finally { if (e_1) throw e_1.error; }
            }
        }
        return rmv;
    };
    return EditableOpLinkedList;
}(OpLinkedList));
exports.EditableOpLinkedList = EditableOpLinkedList;
//# sourceMappingURL=op-linked-list.js.map