"use strict";
/*
    Copyright (C) 2019  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var assert_1 = require("../../util/assert");
var data_validation_1 = require("../../util/data-validation");
var number_1 = require("../../util/number");
var ordering_1 = require("../../util/ordering");
/**
 * @final this class cannot be extended
 *
 * Part of a {@link SimplePos}.
 *
 * A part is a lexicographic quadruple (priority, replica, nth, offset).
 * The triplet (replica, nth, offset) uniquely identifies a part.
 */
var SimplePosPart = /** @class */ (function () {
    /**
     * @param priority {@link SimplePosPart#priority }
     * @param replica {@link SimplePosPart#replica }
     * @param nth {@link SimplePosPart#nth }
     * @param offset {@link SimplePosPart#offset }
     */
    function SimplePosPart(priority, replica, nth, offset) {
        assert_1.assert(function () { return number_1.isU32(priority); }, "random ∈ u32");
        assert_1.assert(function () { return number_1.isU32(replica); }, "replica ∈ u32");
        assert_1.assert(function () { return number_1.isU32(nth); }, "nth ∈ u32");
        assert_1.assert(function () { return number_1.isU32(offset); }, "offset ∈ u32");
        this.priority = priority;
        this.replica = replica;
        this.nth = nth;
        this.offset = offset;
    }
    /**
     * @param priority {@link SimplePosPart#priority }
     *      ∉ {U32_BOTTOM, U32_TOP}
     * @param replica {@link SimplePosPart#replica }
     *      !== U32_TOP
     * @param nth {@link SimplePosPart#nth }
     * @param offset {@link SimplePosPart#offset }
     */
    SimplePosPart.from = function (priority, replica, nth, offset) {
        assert_1.assert(function () { return priority !== number_1.U32_BOTTOM && priority !== number_1.U32_TOP; }, "priority ∉ {U32_BOTTOM, U32_TOP}");
        assert_1.assert(function () { return replica !== number_1.U32_TOP; }, "replica != U32_TOP. This is reserved for BOTTOM and TOP.");
        return new SimplePosPart(priority, replica, nth, offset);
    };
    /**
     * @note {@link SimplePosPart#BOTTOM} and {@link SimplePosPart#TOP}
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x`, or undefined if `x` is not valid.
     */
    SimplePosPart.fromPlain = function (x) {
        if (data_validation_1.isObject(x) &&
            number_1.isU32(x.priority) &&
            number_1.isU32(x.replica) &&
            number_1.isU32(x.nth) &&
            number_1.isU32(x.offset)) {
            return new SimplePosPart(x.priority, x.replica, x.nth, x.offset);
        }
        return undefined;
    };
    // Derivation
    /**
     * @param offset The offset of the new SimplePosPart
     * @return part with the same base, but with a different offset
     */
    SimplePosPart.prototype.withOffset = function (offset) {
        assert_1.assert(function () { return number_1.isU32(offset); }, "offset ∈ u32");
        return new SimplePosPart(this.priority, this.replica, this.nth, offset);
    };
    // Access
    /**
     * @return [priority, replica, nth, offset]
     */
    SimplePosPart.prototype.asTuple = function () {
        return [this.priority, this.replica, this.nth, this.offset];
    };
    /**
     * Non-cryptographic way to approximate object identity.
     */
    SimplePosPart.prototype.hashCode = function () {
        return number_1.hashCodeOf(this.asTuple());
    };
    // Status
    /**
     * @example
     * SimplePosPart(0, 0, 0, _) == SimplePosPart(0, 0, 0, _)
     * SimplePosPart(0, 0, 0, _) < SimplePosPart(1, 0, 0, _)
     * SimplePosPart(1, 0, 0, _) < SimplePosPart(1, 1, 0, _)
     * SimplePosPart(1, 1, 0, _) < SimplePosPart(1, 1, 1, _)
     *
     * Lexicographic order of triplet (priority, replica, nth).
     *
     * @param other
     * @return `this` [Order relation] `other` regardless their offsets
     *  ({@link SimplePosPart#offset})
     */
    SimplePosPart.prototype.compareBase = function (other) {
        return ordering_1.lexCompareOrdering(number_1.compareU32(this.priority, other.priority), ordering_1.lexCompareOrdering(number_1.compareU32(this.replica, other.replica), number_1.compareU32(this.nth, other.nth)));
    };
    /**
     * @example
     * SimplePosPart(0, 0, 0, 0) < SimplePosPart(1, 0, 0, 0)
     * SimplePosPart(1, 0, 0, 0) < SimplePosPart(1, 1, 0, 0)
     * SimplePosPart(1, 1, 0, 0) < SimplePosPart(1, 1, 1, 0)
     * SimplePosPart(1, 1, 1, 0) < SimplePosPart(1, 1, 1, 1)
     *
     * Lexicographic order of quadruple (priority, replica, nth, offset).
     *
     * @param other
     * @return `this` [Order relation] `other`
     */
    SimplePosPart.prototype.compare = function (other) {
        if (this === other) {
            return 0 /* EQUAL */;
        }
        else {
            return ordering_1.lexCompareOrdering(this.compareBase(other), number_1.compareU32(this.offset, other.offset));
        }
    };
    /**
     * Lowest part.
     * Here the triplet (replica, nth, offset) has no meaning.
     */
    SimplePosPart.BOTTOM = new SimplePosPart(number_1.U32_BOTTOM, number_1.U32_TOP, 0, 0);
    /**
     * Greatest part.
     * Here the triplet (replica, nth, offset) has no meaning.
     */
    SimplePosPart.TOP = new SimplePosPart(number_1.U32_TOP, number_1.U32_TOP, 1, 0);
    return SimplePosPart;
}());
exports.SimplePosPart = SimplePosPart;
//# sourceMappingURL=simple-pos-part.js.map