"use strict";
/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var assert_1 = require("../../util/assert");
var data_validation_1 = require("../../util/data-validation");
var number_1 = require("../../util/number");
var pos_1 = require("../../core/pos");
var simple_pos_part_1 = require("./simple-pos-part");
var ordering_1 = require("../../util/ordering");
/**
 * @final this class cannot be extended
 *
 * A position can be represented as a lexicographic list of
 * lexicographic quadruples (priority, replica, nth, offset).
 * Theses triplets are the parts of the position.
 *
 * The triplet (replica, nth, offset) uniquely identifies the position.
 *
 * The set of positions is dense, hence you can always author a new position
 * between two distinct positions. A position can be generated between two
 * int-successive positions (see {@link Pos}) by suffixing a new triplet
 * to the lower position.
 */
var SimplePos = /** @class */ (function () {
    /**
     * @param parts {@link SimplePos#parts }
     */
    function SimplePos(parts) {
        assert_1.assert(function () { return parts.length > 0; }, "parts must not be empty");
        this.parts = parts;
    }
    /**
     * @param parts {@link SimplePos#parts }
     *  The last part must be distinct of SimplePosPart.BOTTOM and
     *  SimplePosPart.TOP.
     * @return Position with {@link parts } as {@link SimplePos#parts }.
     */
    SimplePos.from = function (parts) {
        var lastPart = parts[parts.length - 1];
        assert_1.assert(function () {
            return lastPart.priority !== number_1.U32_BOTTOM &&
                lastPart.priority !== number_1.U32_TOP;
        }, "priority ∉ {U32_BOTTOM, U32_TOP}");
        assert_1.assert(function () { return lastPart.replica !== number_1.U32_TOP; }, "replica != U32_TOP. This is reserved for BOTTOM and TOP.");
        return new SimplePos(parts);
    };
    /**
     * @note {@link SimplePos#BOTTOM } and {@link SimplePos#TOP }
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    SimplePos.fromPlain = function (x) {
        if (data_validation_1.isObject(x) && Array.isArray(x.parts)) {
            var parts = data_validation_1.fromArray(x.parts, simple_pos_part_1.SimplePosPart.fromPlain);
            if (parts !== undefined && parts.length > 0) {
                var lastPart = parts[parts.length - 1];
                if (lastPart.priority !== number_1.U32_BOTTOM &&
                    lastPart.priority !== number_1.U32_TOP &&
                    lastPart.replica !== number_1.U32_TOP) {
                    return SimplePos.from(parts);
                }
            }
        }
        return undefined;
    };
    // Derivation
    /**
     * @param offset The offset of the new position
     * @return Position with the same base, but with a different offset
     */
    SimplePos.prototype.withOffset = function (offset) {
        assert_1.assert(function () { return number_1.isU32(offset); }, "offset ∈ u32");
        var parts = tslib_1.__spread(this.parts);
        var lastIndex = parts.length - 1;
        parts[lastIndex] = parts[lastIndex].withOffset(offset);
        return new SimplePos(parts);
    };
    /**
     * @param n 0-based index.
     * @return `n`-th integer predecessor of this.
     */
    SimplePos.prototype.intPred = function (n) {
        var _this = this;
        assert_1.assert(function () { return number_1.isU32(n); }, "n ∈ u32");
        assert_1.assert(function () { return _this.hasIntPred(n); }, "this has a n-th predecessor");
        return this.withOffset(this.offset() - n);
    };
    /** @override */
    SimplePos.prototype.intSucc = function (n) {
        var _this = this;
        assert_1.assert(function () { return number_1.isU32(n); }, "n ∈ u32");
        assert_1.assert(function () { return _this.hasIntSucc(n); }, "this has a n-th successor");
        return this.withOffset(this.offset() + n);
    };
    // Access
    /**
     * Last part of {@link SimplePosPart#parts }.
     */
    SimplePos.prototype.lastPart = function () {
        return this.parts[this.parts.length - 1];
    };
    /**
     * Number of part in {@link SimplePosPart#parts }.
     */
    SimplePos.prototype.depth = function () {
        return this.parts.length;
    };
    /** @override */
    SimplePos.prototype.replica = function () {
        return this.lastPart().replica;
    };
    /**
     * Each position is generated for the `nth` generated block.
     * The pair (replica, nth) globally and uniquely identifies a block.
     */
    SimplePos.prototype.nth = function () {
        return this.lastPart().nth;
    };
    /**
     * Offset of the position in the block.
     */
    SimplePos.prototype.offset = function () {
        return this.lastPart().offset;
    };
    /** @override */
    SimplePos.prototype.intDistance = function (other) {
        var _this = this;
        if (this.depth() > other.depth()) {
            var _a = tslib_1.__read(other.intDistance(this), 2), dist = _a[0], order = _a[1];
            return [dist, ordering_1.orderingInversion[order]];
        }
        else {
            assert_1.heavyAssert(function () {
                return (function (cmp) {
                    return cmp === -1 /* PREFIXING */ ||
                        cmp === 0 /* EQUAL */;
                })(_this.compareBase(other));
            }, "this prefix or is equal to other");
            var otherIntermediateOffset = other.parts[this.depth() - 1].offset;
            var offset = this.offset();
            return [
                number_1.absoluteSubstraction(offset, otherIntermediateOffset),
                number_1.compareU32(offset, otherIntermediateOffset),
            ];
        }
    };
    /** @override */
    SimplePos.prototype.hashCode = function () {
        return number_1.hashCodeOf(this.parts.map(function (part) { return part.hashCode(); }));
    };
    // Status
    /**
     * hasIntPred(0) is always true.
     *
     * See also {@link Pos#hasIntSucc }.
     *
     * @param n 0-based index.
     * @return is there a {@link n } -th integer predecessor?
     */
    SimplePos.prototype.hasIntPred = function (n) {
        assert_1.assert(function () { return number_1.isU32(n); }, "n ∈ u32");
        return number_1.isU32(this.offset() - n);
    };
    /** @override */
    SimplePos.prototype.hasIntSucc = function (n) {
        assert_1.assert(function () { return number_1.isU32(n); }, "n ∈ u32");
        return number_1.isU32(this.offset() + n);
    };
    /** @override */
    SimplePos.prototype.compareBase = function (other) {
        if (this.depth() > other.depth()) {
            return pos_1.baseOrderingInversion[other.compareBase(this)];
        }
        else if (this.isEqual(other)) {
            return 0 /* EQUAL */;
        }
        else {
            var i = 0;
            var baseCmp = 0 /* EQUAL */;
            while (i < this.depth() - 1 && baseCmp === 0 /* EQUAL */) {
                baseCmp = this.parts[i].compare(other.parts[i]);
                i++;
            }
            if (baseCmp === 0 /* EQUAL */) {
                baseCmp = this.parts[i].compareBase(other.parts[i]);
            }
            switch (baseCmp) {
                case -1 /* BEFORE */:
                    return -2 /* BEFORE */;
                case 0 /* EQUAL */:
                    if (this.depth() === other.depth()) {
                        return 0 /* EQUAL */;
                    }
                    else {
                        return -1 /* PREFIXING */;
                    }
                default:
                    return 2 /* AFTER */;
            }
        }
    };
    /** @override */
    SimplePos.prototype.isBaseEqual = function (other) {
        return this.compareBase(other) === 0 /* EQUAL */;
    };
    /** @override */
    SimplePos.prototype.compare = function (other) {
        if (this.depth() > other.depth()) {
            return ordering_1.orderingInversion[other.compare(this)];
        }
        else if (this.isEqual(other)) {
            return 0 /* EQUAL */;
        }
        else {
            var i = 0;
            var cmp = void 0;
            do {
                cmp = this.parts[i].compare(other.parts[i]);
                i++;
            } while (i < this.depth() && cmp === 0 /* EQUAL */);
            if (cmp === 0 /* EQUAL */) {
                return number_1.compareU32(this.depth(), other.depth());
            }
            else {
                return cmp;
            }
        }
    };
    /** @override */
    SimplePos.prototype.isEqual = function (other) {
        return (this.replica() === other.replica() &&
            this.nth() === other.nth() &&
            this.offset() === other.offset());
    };
    /**
     * Lowest position.
     */
    SimplePos.BOTTOM = new SimplePos([simple_pos_part_1.SimplePosPart.BOTTOM]);
    /**
     * Greatest position.
     */
    SimplePos.TOP = new SimplePos([simple_pos_part_1.SimplePosPart.TOP]);
    return SimplePos;
}());
exports.SimplePos = SimplePos;
//# sourceMappingURL=simple-pos.js.map