/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
import { lexCompareOrdering, compareBoolean } from "../util/ordering";
import { isObject } from "../util/data-validation";
/**
 * An anchor can be used to model the cursor of a participant in a
 * repliacted list.
 */
var Anchor = /** @class */ (function () {
    /**
     * @example
     * Anchor(p) is just before p
     * There not exists another position between Anchor(p) and p
     *
     * @param ref position to which this anchor is relative to
     * @param isAfter is the anchor after `ref`?
     */
    function Anchor(ref, isAfter) {
        this.ref = ref;
        this.isAfter = isAfter;
    }
    /**
     * @example
     * Anchor(p, false) is just before p
     * Anchor(p, true) is just after p
     * There does not exist a position between Anchor(p) and p
     *
     * @param ref position to which this anchor is relative to
     * @param isAfter is the anchor after `ref`?
     */
    Anchor.from = function (ref, isAfter) {
        return new Anchor(ref, isAfter);
    };
    Anchor.fromPlain = function (posFromPlain) {
        return function (x) {
            if (isObject(x) &&
                typeof x.isAfter === "boolean") {
                var ref = posFromPlain(x.ref);
                if (ref !== undefined) {
                    return Anchor.from(ref, x.isAfter);
                }
            }
            return undefined;
        };
    };
    /**
     * Globally unique identifier of the author which generated this block.
     */
    Anchor.prototype.replica = function () {
        return this.ref.replica();
    };
    /**
     * @param other
     * @return this [Order relation] other
     */
    Anchor.prototype.compare = function (other) {
        return lexCompareOrdering(this.ref.compare(other.ref), compareBoolean(other.isAfter, this.isAfter));
    };
    /**
     * @example
     * Anchor(p).compareTo(q) == Ordering.Before if p <= q
     *
     * @param pos
     * @return Is before or after {@link pos}?
     */
    Anchor.prototype.compareWith = function (pos) {
        var rel = this.ref.compare(pos);
        if (rel !== 0 /* EQUAL */) {
            return rel;
        }
        else if (this.isAfter) {
            return 1 /* AFTER */;
        }
        else {
            return -1 /* BEFORE */;
        }
    };
    return Anchor;
}());
export { Anchor };
//# sourceMappingURL=anchor.js.map