import * as tslib_1 from "tslib";
import { isU32 } from "../util/number";
import { assert } from "../util/assert";
import { Ins, Del } from "./local-operation";
import { U32Range } from "./u32-range";
/**
 * Abstraction for block insertion and block removal.
 * The abstraction assume a dichotomic traversal of the list.
 * The abstraction can be implemented by an iterator or a recursive data
 * structure.
 *
 * For example, assume we have the list of blocks [A, C, D, E, F] and we insert
 * the block B (A < B < C < D < E < F).
 * B is first compared to D. Because B < D, we attempt to insert in the
 * left "sub-list" [A, C]. C is compared with A then C or C then A, and is
 * finally inserted between A and C.
 */
var BlockListContext = /** @class */ (function () {
    function BlockListContext() {
    }
    /**
     * Attempt to merge the current block with its predecessor.
     */
    BlockListContext.prototype.mergeLeft = function () {
        var curr = this.current();
        var prev = this.predecessor();
        if (prev !== undefined &&
            prev.compare(curr) === -6 /* PREPENDABLE */) {
            this.removePredecessor();
            this.replace(prev.append(curr));
        }
    };
    /**
     * Attempt to merge the current block with its successor.
     */
    BlockListContext.prototype.mergeRight = function () {
        var curr = this.current();
        var succ = this.successor();
        if (succ !== undefined &&
            succ.compare(curr) === 6 /* APPENDABLE */) {
            this.removeSuccessor();
            this.replace(curr.append(succ));
        }
    };
    /**
     * Remove current block and attempt to merge its predecessor with
     * its successor.
     */
    BlockListContext.prototype.removeCurrentMerge = function () {
        var succ = this.successor();
        if (succ !== undefined) {
            this.replace(succ);
            this.removeSuccessor();
            this.mergeLeft();
        }
        else {
            var prev = this.predecessor();
            if (prev !== undefined) {
                this.replace(prev);
                this.removePredecessor();
            }
            else {
                this.removeCurrent();
            }
        }
    };
    /**
     * @param anchor
     * @param minIndex minimal index in the current sub-list.
     * @return index of `anchor`.
     */
    BlockListContext.prototype.indexFrom = function (anchor, minIndex) {
        assert(function () { return isU32(minIndex); }, "minIndex ∈ u32");
        var curr = this.current();
        var currIndex = minIndex + this.currentRelativeIndex();
        var currEndIndex = currIndex + curr.length;
        var blockIndex = curr.indexFrom(anchor);
        if (blockIndex === 0) {
            return this.indexFromLeft(anchor, minIndex);
        }
        else if (blockIndex === curr.length) {
            return this.indexFromRight(anchor, currEndIndex);
        }
        else {
            return currIndex + blockIndex;
        }
    };
    /**
     * @param relIndex relative index where the anchor is
     * @param isAfter Is the anchor after `relIndex`?
     * @param f factory
     * @return anchor at `relIndex` in the current list.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    BlockListContext.prototype.anchorAt = function (relIndex, isAfter, f) {
        assert(function () { return isU32(relIndex); }, "relIndex ∈ u32");
        var curr = this.current();
        var currRelIndex = this.currentRelativeIndex();
        var currRelEndIndex = currRelIndex + curr.length;
        if (relIndex < currRelIndex) {
            return this.anchorAtLeft(relIndex, isAfter, f);
        }
        else if (relIndex === currRelIndex) {
            if (isAfter) {
                return curr.lowerAnchor();
            }
            else {
                return this.anchorAtLeft(relIndex, isAfter, f);
            }
        }
        else if (relIndex < currRelEndIndex) {
            return curr.anchor(relIndex - currRelIndex, !isAfter);
        }
        else if (relIndex === currRelEndIndex && !isAfter) {
            return curr.upperAnchor();
        }
        else {
            var subRelIndex = relIndex - currRelEndIndex;
            return this.anchorAtRight(subRelIndex, isAfter, f);
        }
    };
    /**
     * @param iBlock
     * @return segments of `iBlock` that can be inserted in the current
     *  sub-list.
     */
    BlockListContext.prototype.insertable = function (iBlock) {
        var curr = this.current();
        switch (iBlock.compare(curr)) {
            case -6 /* PREPENDABLE */:
            case -7 /* BEFORE */:
                return this.insertableLeft(iBlock);
            case 6 /* APPENDABLE */:
            case 7 /* AFTER */:
                return this.insertableRight(iBlock);
            case -5 /* SPLITTED_BY */: {
                var _a = tslib_1.__read(iBlock.splitWith(curr), 2), lSplit = _a[0], rSplit = _a[1];
                var lIns = this.insertableLeft(lSplit);
                var rIns = this.insertableRight(rSplit);
                return lIns.concat(rIns);
            }
            case 5 /* SPLITTING */:
                return [iBlock];
            case 3 /* INCLUDING_RIGHT */:
            case -4 /* OVERLAPPING_BEFORE */:
                return this.insertableLeft(iBlock.prependable(curr));
            case 1 /* INCLUDING_LEFT */:
            case 4 /* OVERLAPPING_AFTER */:
                return this.insertableRight(iBlock.appendable(curr));
            case 2 /* INCLUDING_MIDDLE */: {
                var lIns = this.insertableLeft(iBlock.prependable(curr));
                var rIns = this.insertableRight(iBlock.appendable(curr));
                return lIns.concat(rIns);
            }
            case -2 /* INCLUDED_MIDDLE_BY */:
            case -1 /* INCLUDED_RIGHT_BY */:
            case -3 /* INCLUDED_LEFT_BY */:
            case 0 /* EQUAL */:
                return [];
        }
    };
    /**
     * @param iBlock block to insert.
     * @param minIndex minimal index in the current sub-list.
     * @return Performed modifications in terms of local insertions.
     *  The n+1 -th insertion depends on the effect of the n -th insertion.
     */
    BlockListContext.prototype.insert = function (iBlock, minIndex) {
        assert(function () { return isU32(minIndex); }, "minIndex ∈ u32");
        var curr = this.current();
        var currIndex = minIndex + this.currentRelativeIndex();
        var currEndIndex = currIndex + curr.length;
        switch (iBlock.compare(curr)) {
            case -7 /* BEFORE */:
                return this.insertLeft(iBlock, minIndex);
            case 7 /* AFTER */:
                return this.insertRight(iBlock, currEndIndex);
            case -6 /* PREPENDABLE */: {
                var lIns = this.insertLeft(iBlock, minIndex);
                this.mergeLeft();
                return lIns;
            }
            case 6 /* APPENDABLE */: {
                var rIns = this.insertRight(iBlock, currEndIndex);
                this.mergeRight();
                return rIns;
            }
            case -5 /* SPLITTED_BY */: {
                var _a = tslib_1.__read(iBlock.splitWith(curr), 2), lSplit = _a[0], rSplit = _a[1];
                var rIns = this.insertRight(rSplit, currEndIndex);
                var lIns = this.insertLeft(lSplit, minIndex);
                return rIns.concat(lIns);
            }
            case 5 /* SPLITTING */: {
                var _b = tslib_1.__read(curr.splitWith(iBlock), 2), lSplit = _b[0], rSplit = _b[1];
                this.insertPredecessor(lSplit);
                this.replace(iBlock);
                var iBlockIndex = currIndex + lSplit.length;
                this.insertSuccessor(rSplit);
                return [new Ins(iBlockIndex, iBlock.content)];
            }
            case -4 /* OVERLAPPING_BEFORE */:
            case 4 /* OVERLAPPING_AFTER */:
            case -2 /* INCLUDED_MIDDLE_BY */:
            case -1 /* INCLUDED_RIGHT_BY */:
            case -3 /* INCLUDED_LEFT_BY */:
            case 2 /* INCLUDING_MIDDLE */:
            case 3 /* INCLUDING_RIGHT */:
            case 1 /* INCLUDING_LEFT */:
            case 0 /* EQUAL */:
                throw new Error("inserted block intersects with one or more existing blocks");
        }
    };
    /**
     * @param relIndex relative index
     * @param items elements to insert
     * @param factory [mutated] block factory
     * @return block generated with `factory` that contains `items` as content,
     *  and was inserted at `relIndex` in the current sub-list.
     */
    BlockListContext.prototype.insertAt = function (relIndex, items, factory) {
        assert(function () { return isU32(relIndex); }, "index ∈ u32");
        assert(function () { return items.length > 0; }, "items.length > 0");
        var curr = this.current();
        var currRelIndex = this.currentRelativeIndex();
        var currRelEndIndex = currRelIndex + curr.length;
        if (relIndex < currRelIndex) {
            return this.insertAtLeft(relIndex, items, factory);
        }
        else if (relIndex === currRelIndex) {
            var iBlock = factory.between(this.predecessor(), items, curr);
            this.insert(iBlock, 0); // handle merge
            return iBlock;
        }
        else if (relIndex < currRelEndIndex) {
            var _a = tslib_1.__read(curr.splitAt(relIndex - currRelIndex), 2), lSplit = _a[0], rSplit = _a[1];
            var iBlock = factory.between(lSplit, items, rSplit);
            this.insert(iBlock, 0); // handle split
            return iBlock;
        }
        else if (relIndex === currRelEndIndex) {
            var iBlock = factory.between(curr, items, this.successor());
            this.insert(iBlock, 0); // handle merge
            return iBlock;
        }
        else {
            var subRelIndex = relIndex - currRelEndIndex;
            return this.insertAtRight(subRelIndex, items, factory);
        }
    };
    /**
     * @param dBlock block to remove.
     * @param minIndex minimal index in the current sub-list.
     * @return Performed modifications in terms of local deletions.
     *  The n+1 -th deletion depends on the n -th deletion.
     */
    BlockListContext.prototype.remove = function (dBlock, minIndex) {
        assert(function () { return isU32(minIndex); }, "minIndex ∈ u32");
        var curr = this.current();
        var currIndex = minIndex + this.currentRelativeIndex();
        var currEndIndex = currIndex + curr.length;
        switch (dBlock.compare(curr)) {
            case -6 /* PREPENDABLE */:
            case -7 /* BEFORE */: {
                var lRmv = this.removeLeft(dBlock, minIndex);
                if (lRmv.length > 0 && lRmv[0].endIndex() === currIndex) {
                    this.mergeLeft();
                }
                return lRmv;
            }
            case 6 /* APPENDABLE */:
            case 7 /* AFTER */: {
                var rRmv = this.removeRight(dBlock, currEndIndex);
                var rRmvLen = rRmv.length;
                if (rRmvLen > 0 && rRmv[rRmvLen - 1].index === currEndIndex) {
                    this.mergeRight();
                }
                return rRmv;
            }
            case 1 /* INCLUDING_LEFT */: {
                var rRmv = this.removeRight(dBlock, currEndIndex);
                var currRmv = this.remove(curr.toLengthBlock(), minIndex); // handle merge
                return rRmv.concat(currRmv);
            }
            case 3 /* INCLUDING_RIGHT */: {
                var lRmv = this.removeLeft(dBlock, minIndex);
                var currRmv = this.remove(curr.toLengthBlock(), minIndex); // handle merge
                return currRmv.concat(lRmv);
            }
            case 2 /* INCLUDING_MIDDLE */: {
                var rRmv = this.removeRight(dBlock, currEndIndex);
                var lRmv = this.removeLeft(dBlock, minIndex);
                var currRmv = this.remove(curr.toLengthBlock(), minIndex); // handle merge
                return rRmv.concat(currRmv).concat(lRmv);
            }
            case -4 /* OVERLAPPING_BEFORE */: {
                var removed = dBlock.intersection(curr);
                var rmv = this.remove(removed, minIndex);
                var lRmv = this.removeLeft(dBlock, minIndex);
                return rmv.concat(lRmv);
            }
            case 4 /* OVERLAPPING_AFTER */: {
                var rRmv = this.removeRight(dBlock, currEndIndex);
                var removed = dBlock.intersection(curr);
                var rmv = this.remove(removed, minIndex);
                return rRmv.concat(rmv);
            }
            case -5 /* SPLITTED_BY */: {
                var _a = tslib_1.__read(dBlock.splitWith(curr), 2), lSplit = _a[0], rSplit = _a[1];
                var rRmv = this.remove(rSplit, minIndex); // handle merge
                var lRmv = this.remove(lSplit, minIndex); // handle merge
                return rRmv.concat(lRmv);
            }
            case 0 /* EQUAL */: {
                this.removeCurrentMerge();
                return [new Del(currIndex, curr.length)];
            }
            case -3 /* INCLUDED_LEFT_BY */: {
                this.replace(curr.appendable(dBlock));
                return [new Del(currIndex, dBlock.length)];
            }
            case -1 /* INCLUDED_RIGHT_BY */: {
                var replacing = curr.prependable(dBlock);
                this.replace(replacing);
                return [new Del(currIndex + replacing.length, dBlock.length)];
            }
            case -2 /* INCLUDED_MIDDLE_BY */: {
                var replacing = curr.prependable(dBlock);
                this.replace(replacing);
                this.insertSuccessor(curr.appendable(dBlock));
                return [new Del(currIndex + replacing.length, dBlock.length)];
            }
            case 5 /* SPLITTING */:
                return []; // already removed or not already inserted
        }
    };
    /**
     * @param dRange range of indexes to remove
     * @param minIndex minimal index in the current sub-list.
     * @return segments of removed block
     */
    BlockListContext.prototype.removeAt = function (dRange, minIndex) {
        assert(function () { return isU32(minIndex); }, "minIndex ∈ u32");
        var curr = this.current();
        var currIndex = minIndex + this.currentRelativeIndex();
        var currEndIndex = currIndex + curr.length;
        var currRange = U32Range.fromLength(currIndex, curr.length);
        switch (dRange.compare(currRange)) {
            case -6 /* BEFORE */:
                return this.removeAtLeft(dRange, minIndex);
            case 6 /* AFTER */:
                return this.removeAtRight(dRange, currEndIndex);
            case -5 /* PREPENDABLE */: {
                var lRmv = this.removeAtLeft(dRange, minIndex);
                this.mergeLeft();
                return lRmv;
            }
            case 5 /* APPENDABLE */: {
                var rRmv = this.removeAtRight(dRange, currEndIndex);
                this.mergeRight();
                return rRmv;
            }
            case -4 /* OVERLAPPING_BEFORE */: {
                var rmvRange = dRange.intersection(currRange);
                var currRmv = this.removeAt(rmvRange, minIndex);
                var lRmv = this.removeAtLeft(dRange, minIndex);
                return lRmv.concat(currRmv);
            }
            case 4 /* OVERLAPPING_AFTER */: {
                var rRmv = this.removeAtRight(dRange, currEndIndex);
                var rmvRange = dRange.intersection(currRange);
                var currRmv = this.removeAt(rmvRange, minIndex);
                return currRmv.concat(rRmv);
            }
            case 0 /* EQUAL */: {
                this.removeCurrentMerge();
                return [curr.toLengthBlock()];
            }
            case 1 /* INCLUDING_LEFT */: {
                var rRmv = this.removeAtRight(dRange, currEndIndex);
                this.removeCurrentMerge();
                return [curr.toLengthBlock()].concat(rRmv);
            }
            case 3 /* INCLUDING_RIGHT */: {
                var lRmv = this.removeAtLeft(dRange, minIndex);
                this.removeCurrentMerge();
                return lRmv.concat([curr.toLengthBlock()]);
            }
            case 2 /* INCLUDING_MIDDLE */: {
                var rRmv = this.removeAtRight(dRange, currEndIndex);
                var lRmv = this.removeAtLeft(dRange, minIndex);
                this.removeCurrentMerge();
                return lRmv.concat([curr.toLengthBlock()]).concat(rRmv);
            }
            case -3 /* INCLUDED_LEFT_BY */: {
                var _a = tslib_1.__read(curr.splitAt(dRange.length), 2), lSplit = _a[0], rSplit = _a[1];
                this.replace(rSplit);
                return [lSplit.toLengthBlock()];
            }
            case -1 /* INCLUDED_RIGHT_BY */: {
                var splitIndex = dRange.lower - currRange.lower;
                var _b = tslib_1.__read(curr.splitAt(splitIndex), 2), lSplit = _b[0], rSplit = _b[1];
                this.replace(lSplit);
                return [rSplit.toLengthBlock()];
            }
            case -2 /* INCLUDED_MIDDLE_BY */: {
                var rSplitIndex = dRange.upper() + 1 - currRange.lower;
                var _c = tslib_1.__read(curr.splitAt(rSplitIndex), 2), lSplit = _c[0], rSplit = _c[1];
                var lSplitIndex = dRange.lower - currRange.lower;
                var _d = tslib_1.__read(lSplit.splitAt(lSplitIndex), 2), llSplit = _d[0], lrSplit = _d[1];
                this.replace(llSplit);
                this.insertSuccessor(rSplit);
                return [lrSplit.toLengthBlock()];
            }
        }
    };
    return BlockListContext;
}());
export { BlockListContext };
//# sourceMappingURL=block-list-context.js.map