/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
var _a;
import * as tslib_1 from "tslib";
import { assert, heavyAssert } from "../util/assert";
import { isObject } from "../util/data-validation";
import { hashCodeOf, isU32 } from "../util/number";
import { Anchor } from "./anchor";
import { ConcatLength } from "./concat";
import { U32Range } from "./u32-range";
/**
 * Map IntervalOrdering to BlockOrdering.
 */
var rangeOrderingAsBlockOrdering = (_a = {},
    _a[-6 /* BEFORE */] = -7 /* BEFORE */,
    _a[-5 /* PREPENDABLE */] = -6 /* PREPENDABLE */,
    _a[-4 /* OVERLAPPING_BEFORE */] = -4 /* OVERLAPPING_BEFORE */,
    _a[1 /* INCLUDING_LEFT */] = 1 /* INCLUDING_LEFT */,
    _a[2 /* INCLUDING_MIDDLE */] = 2 /* INCLUDING_MIDDLE */,
    _a[3 /* INCLUDING_RIGHT */] = 3 /* INCLUDING_RIGHT */,
    _a[0 /* EQUAL */] = 0 /* EQUAL */,
    _a[-3 /* INCLUDED_LEFT_BY */] = -3 /* INCLUDED_LEFT_BY */,
    _a[-2 /* INCLUDED_MIDDLE_BY */] = -2 /* INCLUDED_MIDDLE_BY */,
    _a[-1 /* INCLUDED_RIGHT_BY */] = -1 /* INCLUDED_RIGHT_BY */,
    _a[4 /* OVERLAPPING_AFTER */] = 4 /* OVERLAPPING_AFTER */,
    _a[5 /* APPENDABLE */] = 6 /* APPENDABLE */,
    _a[6 /* AFTER */] = 7 /* AFTER */,
    _a);
/**
 * Segment of concatenable elements generated by the same author.
 *
 * A segment can be represented as a set of int successive positions
 * (see {@link Pos }) and an attached content.
 */
var Block = /** @class */ (function () {
    /**
     * @param lowerPos First position of the block.
     * @param content Content of the block.
     */
    function Block(lowerPos, content) {
        this.lowerPos = lowerPos;
        this.content = content;
        assert(function () { return content.length > 0; }, "content.length > 0");
        assert(function () { return lowerPos.hasIntSucc(content.length - 1); }, "valid positions");
    }
    /**
     * @param posFromPlain
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a block.
     *  It returns the built block if it succeeds, or undefined if it fails.
     */
    Block.fromPlain = function (posFromPlain, itemsFromPlain) {
        return function (x) {
            if (isObject(x)) {
                var pos = posFromPlain(x.lowerPos);
                var content = itemsFromPlain(x.content);
                if (pos !== undefined && content !== undefined) {
                    return new Block(pos, content);
                }
            }
            return undefined;
        };
    };
    Object.defineProperty(Block.prototype, "length", {
        // Access
        /**
         * Length of {@link Block#content }
         */
        get: function () {
            return this.content.length;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @example
     * b.nthPos(0) == b.lowerPos
     * b.nthPos(b.length - 1) == b.upperPos
     *
     * @param nth 0-based index
     * @return {@link nth } -th position of the block.
     */
    Block.prototype.nthPos = function (nth) {
        var _this = this;
        assert(function () { return isU32(nth); }, "nth ∈ u32");
        assert(function () { return nth < _this.length; }, "nth < this.length");
        if (nth === 0) {
            return this.lowerPos;
        }
        else {
            return this.lowerPos.intSucc(nth);
        }
    };
    /**
     * Last position of the block.
     */
    Block.prototype.upperPos = function () {
        return this.nthPos(this.length - 1);
    };
    /**
     * @return anchor just before the lower position
     */
    Block.prototype.lowerAnchor = function () {
        return this.anchor(0, false);
    };
    /**
     * @param nth 0-based index
     * @param isAfter is the anchor after the `nth` position?
     * @return Anchor relative to the `nth` psoyion of the block.
     */
    Block.prototype.anchor = function (nth, isAfter) {
        var _this = this;
        assert(function () { return isU32(nth); }, "nth ∈ u32");
        assert(function () { return nth < _this.length; }, "nth < this.length");
        return Anchor.from(this.nthPos(nth), isAfter);
    };
    /**
     * @return anchor just after the upper position
     */
    Block.prototype.upperAnchor = function () {
        return this.anchor(this.length - 1, true);
    };
    /**
     * Globally unique identifier of the author which generated this block.
     */
    Block.prototype.replica = function () {
        return this.lowerPos.replica();
    };
    /**
     * When each position of this block were generated.
     */
    Block.prototype.seqs = function () {
        return U32Range.fromLength(this.lowerPos.seq(), this.length);
    };
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return 0-based index where {@link other } splits this block.
     */
    Block.prototype.splittingIndex = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.compare(other) === -5 /* SPLITTED_BY */; }, "other splits this");
        var _a = tslib_1.__read(this.lowerPos.intDistance(other.lowerPos), 1), dist = _a[0];
        return dist + 1;
    };
    /**
     * @return Length block representation of this block.
     */
    Block.prototype.toLengthBlock = function () {
        var concatLength = new ConcatLength(this.length);
        return new Block(this.lowerPos, concatLength);
    };
    /**
     * Non-cryptographic way to approximate object identity.
     * Do not take the content into account.
     */
    Block.prototype.structuralHashCode = function () {
        return hashCodeOf([this.lowerPos.hashCode(), this.length]);
    };
    // Status
    /**
     * @return Is this block a length block?
     */
    Block.prototype.isLengthBlock = function () {
        return this.content instanceof ConcatLength;
    };
    /**
     * @param other
     * @return Are this and {@link other } an intersection?
     */
    Block.prototype.hasIntersection = function (other) {
        var cmp = this.compare(other);
        return (-4 /* OVERLAPPING_BEFORE */ <= cmp &&
            cmp <= 4 /* OVERLAPPING_AFTER */);
    };
    /**
     * @param other
     * @return Has this an appendable segment to {@link other }?
     */
    Block.prototype.hasAppendable = function (other) {
        var cmp = this.compare(other);
        return (cmp === 4 /* OVERLAPPING_AFTER */ ||
            cmp === 1 /* INCLUDING_LEFT */ ||
            cmp === 2 /* INCLUDING_MIDDLE */);
    };
    /**
     * @param other
     * @return Has this a prependable segment to {@link other }?
     */
    Block.prototype.hasPrependable = function (other) {
        var cmp = this.compare(other);
        return (cmp === -4 /* OVERLAPPING_BEFORE */ ||
            cmp === 3 /* INCLUDING_RIGHT */ ||
            cmp === 2 /* INCLUDING_MIDDLE */);
    };
    /**
     * @param anchor
     * @return A pair that includes the index of `anchor` in the block and
     *  its order relation.
     *  Ordering.EQUAL means that {@link anchor} is realted to a position
     *  of this block.
     */
    Block.prototype.indexFrom = function (anchor) {
        var baseCmp = this.lowerPos.compareBase(anchor.ref);
        if (baseCmp === -2 /* BEFORE */) {
            return this.length;
        }
        else if (baseCmp === 2 /* AFTER */) {
            return 0;
        }
        else {
            var _a = tslib_1.__read(this.lowerPos.intDistance(anchor.ref), 2), dist = _a[0], order = _a[1];
            if (order === -1 /* BEFORE */) {
                return this.length;
            }
            else if (dist >= this.length) {
                return 0;
            }
            else {
                if (anchor.isAfter) {
                    return dist + 1;
                }
                else {
                    return dist;
                }
            }
        }
    };
    /**
     * @example
     * a.compare(b) == Ordering.Before if a < b
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    Block.prototype.compare = function (other) {
        var baseCmp = this.lowerPos.compareBase(other.lowerPos);
        if (baseCmp === -2 /* BEFORE */) {
            return -7 /* BEFORE */; // Before meta block
        }
        else if (baseCmp === 2 /* AFTER */) {
            return 7 /* AFTER */; // After meta block
        }
        else {
            var _a = tslib_1.__read(this.lowerPos.intDistance(other.lowerPos), 2), dist = _a[0], order = _a[1];
            if (baseCmp === 0 /* EQUAL */) {
                var lower1 = order === -1 /* BEFORE */ ? 0 : dist;
                var thisInterval = U32Range.fromLength(lower1, this.length);
                var lower2 = order === -1 /* BEFORE */ ? dist : 0;
                var otherInterval = U32Range.fromLength(lower2, other.length);
                var intervalCmp = thisInterval.compare(otherInterval);
                return rangeOrderingAsBlockOrdering[intervalCmp];
            }
            else if (baseCmp === -1 /* PREFIXING */) {
                if (order === 1 /* AFTER */) {
                    return 7 /* AFTER */;
                }
                else if (dist >= this.length - 1) {
                    return -7 /* BEFORE */;
                }
                else {
                    return -5 /* SPLITTED_BY */;
                }
            }
            else {
                if (order === -1 /* BEFORE */) {
                    return -7 /* BEFORE */;
                }
                else if (dist >= other.length - 1) {
                    return 7 /* AFTER */;
                }
                else {
                    return 5 /* SPLITTING */;
                }
            }
        }
    };
    // Derivation
    /**
     * @example
     * Block(p, "ab").append(Block(_, "cd")) == Block(p, "abcd")
     *
     * @param other
     *      {@link other } is appendable to this.
     * @return other appended to this.
     */
    Block.prototype.append = function (other) {
        var _this = this;
        heavyAssert(function () { return other.compare(_this) === 6 /* APPENDABLE */; }, "other is appendable to this");
        return new Block(this.lowerPos, this.content.concat(other.content));
    };
    /**
     * @example
     * Block(p, "ab").leftSplitAt(1) == Block(p, "a")
     * Block(p, "ab").leftSplitAt(2) == Block(p, "ab")
     *
     * @param index 0-based index
     *      0 < {@link index } <= {@link Block#length }
     * @return Left split.
     */
    Block.prototype.leftSplitAt = function (index) {
        var _this = this;
        assert(function () { return isU32(index); }, "index ∈ u32");
        assert(function () { return 0 < index && index <= _this.length; }, "0 < index <= this.length");
        var leftItems = this.content.slice(0, index);
        return new Block(this.lowerPos, leftItems);
    };
    /**
     * @example
     * Block(p, "ab").rightSplitAt(1) == Block(p+1, "b")
     * Block(p, "ab").rightSplitAt(0) == Block(p, "ab")
     *
     * @param index 0-based index
     *      {@link index } < this.length
     * @return Right split.
     */
    Block.prototype.rightSplitAt = function (index) {
        var _this = this;
        assert(function () { return isU32(index); }, "index ∈ u32");
        assert(function () { return index < _this.length; }, "0 <= index < this.length");
        var rightItems = this.content.slice(index, this.length);
        var rightPos = this.nthPos(index);
        return new Block(rightPos, rightItems);
    };
    /**
     * @example
     * b.splitAt(n) == [b.leftSplitAt(n), b.rightSplitAt(n)]
     *
     * @param index 0-based index
     *      0 < {@link index } < {@link Block#length }
     * @return Left and right splits.
     */
    Block.prototype.splitAt = function (index) {
        var _this = this;
        assert(function () { return isU32(index); }, "index ∈ u32");
        assert(function () { return 0 < index && index < _this.length; }, "0 < index < this.length");
        return [this.leftSplitAt(index), this.rightSplitAt(index)];
    };
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return Left and right splits.
     */
    Block.prototype.splitWith = function (other) {
        var _this = this;
        assert(function () { return _this.compare(other) === -5 /* SPLITTED_BY */; }, "other splits this");
        return this.splitAt(this.splittingIndex(other));
    };
    /**
     * @param other
     *      this includes or overlaps after {@link other }
     * @return Part of this block which can be append to {@link block }.
     */
    Block.prototype.appendable = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.hasAppendable(other); }, "this has an appendable segment to other");
        var _a = tslib_1.__read(this.lowerPos.intDistance(other.lowerPos), 2), dist = _a[0], order = _a[1];
        if (order === -1 /* BEFORE */) {
            return this.rightSplitAt(other.length + dist);
        }
        else {
            return this.rightSplitAt(other.length - dist);
        }
    };
    /**
     * @param other
     *      this includes or overlaps before {@link other }
     * @return Part of this block which can be prepend to {@link block }.
     */
    Block.prototype.prependable = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.hasPrependable(other); }, "this has a prependable segment to other");
        var _a = tslib_1.__read(this.lowerPos.intDistance(other.lowerPos), 1), dist = _a[0];
        return this.leftSplitAt(dist);
    };
    /**
     * @param other
     *      this and {@link other } intersect.
     * @return Intersection part between this and {@link other }.
     */
    Block.prototype.intersection = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.hasIntersection(other); }, "this intersects other.");
        var _a = tslib_1.__read(this.lowerPos.intDistance(other.lowerPos), 2), dist = _a[0], order = _a[1];
        if (order === -1 /* BEFORE */) {
            var pos = other.lowerPos;
            var l = Math.min(this.length - dist, other.length);
            var elements = this.content.slice(dist, dist + l);
            return new Block(pos, elements);
        }
        else {
            var pos = this.lowerPos;
            var l = Math.min(other.length - dist, this.length);
            var elements = this.content.slice(0, l);
            return new Block(pos, elements);
        }
    };
    return Block;
}());
export { Block };
//# sourceMappingURL=block.js.map