/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
import * as tslib_1 from "tslib";
import { assert, heavyAssert } from "../../util/assert";
import { isObject, fromArray } from "../../util/data-validation";
import { absoluteSubstraction, compareU32, hashCodeOf, isU32, U32_BOTTOM, U32_TOP, } from "../../util/number";
import { baseOrderingInversion } from "../../core/pos";
import { SimpleDotPosPart } from "./simple-dot-pos-part";
import { orderingInversion } from "../../util/ordering";
/**
 * @final this class cannot be extended
 *
 * A position can be represented as a lexicographic list of
 * lexicographic triplets (priority, replica, seq).
 * Theses triplets are the parts of the position.
 *
 * The dot (replica, seq) uniquely identifies the position.
 *
 * The set of positions is dense, hence you can always author a new position
 * between two distinct positions. A position can be generated between two
 * int-successive positions (see {@link Pos}) by suffixing a new triplet
 * to the lower position.
 */
var SimpleDotPos = /** @class */ (function () {
    /**
     * @param parts {@link SimpleDotPos#parts }
     */
    function SimpleDotPos(parts) {
        assert(function () { return parts.length > 0; }, "parts must not be empty");
        this.parts = parts;
    }
    /**
     * @param parts {@link SimpleDotPos#parts }
     *  The last part must be distinct of SimplePosPart.BOTTOM and
     *  SimplePosPart.TOP.
     * @return Position with {@link parts } as {@link SimpleDotPos#parts }.
     */
    SimpleDotPos.from = function (parts) {
        var lastPart = parts[parts.length - 1];
        assert(function () {
            return lastPart.priority !== U32_BOTTOM &&
                lastPart.priority !== U32_TOP;
        }, "priority ∉ {U32_BOTTOM, U32_TOP}");
        assert(function () { return lastPart.replica !== U32_TOP; }, "replica != U32_TOP. Reserved for BOTTOM and TOP.");
        return new SimpleDotPos(parts);
    };
    /**
     * @note {@link SimpleDotPos#BOTTOM } and {@link SimpleDotPos#TOP }
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    SimpleDotPos.fromPlain = function (x) {
        if (isObject(x) && Array.isArray(x.parts)) {
            var parts = fromArray(x.parts, SimpleDotPosPart.fromPlain);
            if (parts !== undefined && parts.length > 0) {
                var lastPart = parts[parts.length - 1];
                if (lastPart.priority !== U32_BOTTOM &&
                    lastPart.priority !== U32_TOP &&
                    lastPart.replica !== U32_TOP) {
                    return SimpleDotPos.from(parts);
                }
            }
        }
        return undefined;
    };
    // Derivation
    /**
     * @param seq The seq of the new position
     * @return Position with the same base, but with a different seq.
     */
    SimpleDotPos.prototype.withSeq = function (seq) {
        assert(function () { return isU32(seq); }, "seq ∈ u32");
        assert(function () { return seq !== 0; }, "seq != 0. Reserved for convenience");
        var parts = tslib_1.__spread(this.parts);
        var lastIndex = parts.length - 1;
        parts[lastIndex] = parts[lastIndex].withSeq(seq);
        return new SimpleDotPos(parts);
    };
    /** @override */
    SimpleDotPos.prototype.intSucc = function (n) {
        var _this = this;
        assert(function () { return isU32(n); }, "n ∈ u32");
        assert(function () { return _this.hasIntSucc(n); }, "this has a n-th successor");
        return this.withSeq(this.seq() + n);
    };
    /**
     * Last part of {@link SimplePosPart#parts }.
     */
    SimpleDotPos.prototype.lastPart = function () {
        return this.parts[this.parts.length - 1];
    };
    /**
     * Number of part in {@link SimplePosPart#parts }.
     */
    SimpleDotPos.prototype.depth = function () {
        return this.parts.length;
    };
    /** @override */
    SimpleDotPos.prototype.replica = function () {
        return this.lastPart().replica;
    };
    /** @override */
    SimpleDotPos.prototype.seq = function () {
        return this.lastPart().seq;
    };
    /** @override */
    SimpleDotPos.prototype.intDistance = function (other) {
        var _this = this;
        if (this.depth() > other.depth()) {
            var _a = tslib_1.__read(other.intDistance(this), 2), dist = _a[0], order = _a[1];
            return [dist, orderingInversion[order]];
        }
        else {
            heavyAssert(function () {
                return (function (cmp) {
                    return cmp === -1 /* PREFIXING */ ||
                        cmp === 0 /* EQUAL */;
                })(_this.compareBase(other));
            }, "this prefix or is equal to other");
            var otherIntermediateSeq = other.parts[this.depth() - 1].seq;
            var seq = this.seq();
            return [
                absoluteSubstraction(seq, otherIntermediateSeq),
                compareU32(seq, otherIntermediateSeq),
            ];
        }
    };
    /** @override */
    SimpleDotPos.prototype.hashCode = function () {
        return hashCodeOf(this.parts.map(function (part) { return part.hashCode(); }));
    };
    // Status
    /** @override */
    SimpleDotPos.prototype.hasIntSucc = function (n) {
        assert(function () { return isU32(n); }, "n ∈ u32");
        return isU32(this.seq() + n);
    };
    /** @override */
    SimpleDotPos.prototype.compareBase = function (other) {
        if (this.depth() > other.depth()) {
            return baseOrderingInversion[other.compareBase(this)];
        }
        else if (this.isEqual(other)) {
            return 0 /* EQUAL */;
        }
        else {
            var i = 0;
            var baseCmp = 0 /* EQUAL */;
            while (i < this.depth() - 1 && baseCmp === 0 /* EQUAL */) {
                baseCmp = this.parts[i].compare(other.parts[i]);
                i++;
            }
            if (baseCmp === 0 /* EQUAL */) {
                baseCmp = this.parts[i].compareBase(other.parts[i]);
            }
            switch (baseCmp) {
                case -1 /* BEFORE */:
                    return -2 /* BEFORE */;
                case 0 /* EQUAL */:
                    if (this.depth() === other.depth()) {
                        return 0 /* EQUAL */;
                    }
                    else {
                        return -1 /* PREFIXING */;
                    }
                default:
                    return 2 /* AFTER */;
            }
        }
    };
    /** @override */
    SimpleDotPos.prototype.isBaseEqual = function (other) {
        return this.compareBase(other) === 0 /* EQUAL */;
    };
    /** @override */
    SimpleDotPos.prototype.compare = function (other) {
        if (this.depth() > other.depth()) {
            return orderingInversion[other.compare(this)];
        }
        else if (this.isEqual(other)) {
            return 0 /* EQUAL */;
        }
        else {
            var i = 0;
            var cmp = void 0;
            do {
                cmp = this.parts[i].compare(other.parts[i]);
                i++;
            } while (i < this.depth() && cmp === 0 /* EQUAL */);
            if (cmp === 0 /* EQUAL */) {
                return compareU32(this.depth(), other.depth());
            }
            else {
                return cmp;
            }
        }
    };
    /** @override */
    SimpleDotPos.prototype.isEqual = function (other) {
        return this.replica() === other.replica() && this.seq() === other.seq();
    };
    /**
     * Lowest position.
     */
    SimpleDotPos.BOTTOM = new SimpleDotPos([SimpleDotPosPart.BOTTOM]);
    /**
     * Greatest position.
     */
    SimpleDotPos.TOP = new SimpleDotPos([SimpleDotPosPart.TOP]);
    return SimpleDotPos;
}());
export { SimpleDotPos };
//# sourceMappingURL=simple-dot-pos.js.map