import * as tslib_1 from "tslib";
import { OpReplicatedList, } from "../../core/op-replicated-list";
import { ValuedNode, lengthOf } from "./avl-list-node";
import { isU32 } from "../../util/number";
import { Ins } from "../../core/local-operation";
import { assert } from "../../util/assert";
import { U32Range } from "../../core/u32-range";
import { isObject } from "../../util/data-validation";
/**
 * An {@see OpReplicatedList } that uses an AVL tree.
 */
var OpAvlList = /** @class */ (function (_super) {
    tslib_1.__extends(OpAvlList, _super);
    /**
     * @param root {@see EditableOpAvlList#root }
     */
    function OpAvlList(root) {
        var _this = _super.call(this) || this;
        _this.root = root;
        return _this;
    }
    /**
     * @param v empty value of type E
     * @return new empty list.
     */
    OpAvlList.empty = function () {
        return new OpAvlList(undefined);
    };
    /**
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    OpAvlList.fromPlain = function (f, itemsFromPlain) {
        return function (x) {
            if (isObject(x)) {
                var blockFromPlain = f.blockFromPlain(itemsFromPlain);
                var root = ValuedNode.fromPlain(blockFromPlain)(x.root);
                return new OpAvlList(root);
            }
            return undefined;
        };
    };
    Object.defineProperty(OpAvlList.prototype, "length", {
        /** @Override */
        get: function () {
            return lengthOf(this.root);
        },
        enumerable: true,
        configurable: true
    });
    /** @Override */
    OpAvlList.prototype.indexFrom = function (anchor) {
        if (this.root !== undefined) {
            return this.root.indexFrom(anchor, 0);
        }
        else {
            return 0;
        }
    };
    /** @Override */
    OpAvlList.prototype.reduceBlock = function (f, prefix) {
        if (this.root !== undefined) {
            return this.root.reduceBlock(f, prefix);
        }
        return prefix;
    };
    /** @Override */
    OpAvlList.prototype.insertable = function (delta) {
        if (this.root !== undefined) {
            return this.root.insertable(delta);
        }
        return [];
    };
    // Modification
    /** @Override */
    OpAvlList.prototype.insert = function (delta) {
        if (this.root !== undefined) {
            var ins = this.root.insert(delta, 0);
            this.root = this.root.balance();
            return ins;
        }
        else {
            this.root = ValuedNode.leaf(delta);
            return [new Ins(0, delta.content)];
        }
    };
    /** @Override */
    OpAvlList.prototype.remove = function (delta) {
        if (this.root !== undefined) {
            var rmv = this.root.remove(delta, 0);
            this.root = this.root.updated();
            return rmv;
        }
        return [];
    };
    return OpAvlList;
}(OpReplicatedList));
export { OpAvlList };
/**
 * An {@see EditableOpReplicatedList } that uses an AVL tree.
 */
var EditableOpAvlList = /** @class */ (function (_super) {
    tslib_1.__extends(EditableOpAvlList, _super);
    /**
     * @param root {@see EditableOpAvlList#root }
     * @param factory {@see EditableOpAvlList#factory }
     */
    function EditableOpAvlList(root, factory) {
        var _this = _super.call(this, root) || this;
        _this.factory = factory;
        return _this;
    }
    /**
     * @param factory factory of blocks
     * @param v empty value of type E (only used for type inference)
     * @return new empty list.
     */
    EditableOpAvlList.emptyWith = function (factory) {
        return new EditableOpAvlList(undefined, factory);
    };
    /**
     *
     * @param f
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    EditableOpAvlList.fromPlain = function (f, itemsFromPlain) {
        return function (x) {
            if (isObject(x)) {
                var factory = f.fromPlain(x.factory);
                if (factory !== undefined) {
                    var blockFromPlain = f.blockFromPlain(itemsFromPlain);
                    var root = ValuedNode.fromPlain(blockFromPlain)(x.root);
                    return new EditableOpAvlList(root, factory);
                }
            }
            return undefined;
        };
    };
    /** @Override */
    EditableOpAvlList.prototype.anchorAt = function (index, isAfter) {
        assert(function () { return isU32(index); }, "index ∈ u32");
        if (this.root !== undefined) {
            return this.root.anchorAt(index, isAfter, this.factory);
        }
        else if (isAfter) {
            return this.factory.topAnchor;
        }
        else {
            return this.factory.bottomAnchor;
        }
    };
    /** @Override */
    EditableOpAvlList.prototype.remove = function (delta) {
        this.factory.garbageCollect(delta);
        return _super.prototype.remove.call(this, delta);
    };
    /** @Override */
    EditableOpAvlList.prototype.insertAt = function (index, items) {
        var _this = this;
        assert(function () { return isU32(index); }, "index ∈ u32");
        assert(function () { return index <= _this.length; }, "valid index");
        assert(function () { return items.length > 0; }, "items is not empty");
        if (this.root !== undefined) {
            var iBlock = this.root.insertAt(index, items, this.factory);
            this.root = this.root.balance();
            return iBlock;
        }
        else {
            var iBlock = this.factory.from(items);
            this.root = ValuedNode.leaf(iBlock);
            return iBlock;
        }
    };
    /** @Override */
    EditableOpAvlList.prototype.removeAt = function (index, length) {
        var e_1, _a;
        var _this = this;
        assert(function () { return isU32(index); }, "index ∈ u32");
        assert(function () { return index < _this.length; }, "valid index");
        assert(function () { return isU32(length); }, "length ∈ u32");
        assert(function () { return index + length <= _this.length; }, "valid end index");
        if (this.root !== undefined) {
            var indexRange = U32Range.fromLength(index, length);
            var rmv = this.root.removeAt(indexRange, 0);
            this.root = this.root.updated();
            if (this.factory.canGarbageCollect()) {
                try {
                    for (var rmv_1 = tslib_1.__values(rmv), rmv_1_1 = rmv_1.next(); !rmv_1_1.done; rmv_1_1 = rmv_1.next()) {
                        var dBlock = rmv_1_1.value;
                        this.factory.garbageCollect(dBlock);
                    }
                }
                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                finally {
                    try {
                        if (rmv_1_1 && !rmv_1_1.done && (_a = rmv_1.return)) _a.call(rmv_1);
                    }
                    finally { if (e_1) throw e_1.error; }
                }
            }
            return rmv;
        }
        return [];
    };
    return EditableOpAvlList;
}(OpAvlList));
export { EditableOpAvlList };
//# sourceMappingURL=op-avl-list.js.map