import { FromPlain } from "../util/data-validation";
import { u32 } from "../util/number";
import { Anchor } from "./anchor";
import { Pos } from "./pos";
import { Concat, ConcatLength } from "./concat";
import { U32Range } from "./u32-range";
import { DotPos } from "./dot-pos";
/**
 * Possible relation between two blocks.
 *
 * Same semantic as {@link IntervalOrdering }.
 * @example
 * [0, 2] |⊂ 1[_, _]
 */
export declare const enum BlockOrdering {
    BEFORE = -7,
    PREPENDABLE = -6,
    SPLITTED_BY = -5,
    OVERLAPPING_BEFORE = -4,
    INCLUDED_LEFT_BY = -3,
    INCLUDED_MIDDLE_BY = -2,
    INCLUDED_RIGHT_BY = -1,
    EQUAL = 0,
    INCLUDING_LEFT = 1,
    INCLUDING_MIDDLE = 2,
    INCLUDING_RIGHT = 3,
    OVERLAPPING_AFTER = 4,
    SPLITTING = 5,
    APPENDABLE = 6,
    AFTER = 7
}
/**
 * A {@see Block} or a {@see LengthBlock}.
 */
export interface BaseBlockk<P extends Pos<P>> {
    /**
     * First position of the block.
     */
    readonly lowerPos: P;
    /**
     * Last position of the block.
     */
    readonly upperPos: () => P;
    /**
     * Block length.
     */
    readonly length: u32;
}
/**
 * Represent the removal of a segment.
 */
export declare type LengthBlock<P extends Pos<P>> = Block<P, ConcatLength>;
/**
 * Segment of concatenable elements generated by the same author.
 *
 * A segment can be represented as a set of int successive positions
 * (see {@link Pos }) and an attached content.
 */
export declare class Block<P extends Pos<P>, E extends Concat<E>> {
    readonly lowerPos: P;
    readonly content: E;
    /**
     * @param lowerPos First position of the block.
     * @param content Content of the block.
     */
    constructor(lowerPos: P, content: E);
    /**
     * @param posFromPlain
     * @param itemsFromPlain
     * @return function that accepts a value and attempt to build a block.
     *  It returns the built block if it succeeds, or undefined if it fails.
     */
    static fromPlain<P extends Pos<P>, E extends Concat<E>>(posFromPlain: FromPlain<P>, itemsFromPlain: FromPlain<E>): FromPlain<Block<P, E>>;
    /**
     * Length of {@link Block#content }
     */
    readonly length: u32;
    /**
     * @example
     * b.nthPos(0) == b.lowerPos
     * b.nthPos(b.length - 1) == b.upperPos
     *
     * @param nth 0-based index
     * @return {@link nth } -th position of the block.
     */
    nthPos(nth: u32): P;
    /**
     * Last position of the block.
     */
    upperPos(): P;
    /**
     * @return anchor just before the lower position
     */
    lowerAnchor(): Anchor<P>;
    /**
     * @param nth 0-based index
     * @param isAfter is the anchor after the `nth` position?
     * @return Anchor relative to the `nth` psoyion of the block.
     */
    anchor(nth: u32, isAfter: boolean): Anchor<P>;
    /**
     * @return anchor just after the upper position
     */
    upperAnchor(): Anchor<P>;
    /**
     * Globally unique identifier of the author which generated this block.
     */
    replica(this: BaseBlockk<P>): ReturnType<P["replica"]>;
    /**
     * When each position of this block were generated.
     */
    seqs<P extends DotPos<P>>(this: BaseBlockk<P>): U32Range;
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return 0-based index where {@link other } splits this block.
     */
    splittingIndex(other: BaseBlockk<P>): u32;
    /**
     * @return Length block representation of this block.
     */
    toLengthBlock(): LengthBlock<P>;
    /**
     * Non-cryptographic way to approximate object identity.
     * Do not take the content into account.
     */
    structuralHashCode(): u32;
    /**
     * @return Is this block a length block?
     */
    isLengthBlock(): this is LengthBlock<P>;
    /**
     * @param other
     * @return Are this and {@link other } an intersection?
     */
    hasIntersection(other: BaseBlockk<P>): boolean;
    /**
     * @param other
     * @return Has this an appendable segment to {@link other }?
     */
    hasAppendable(other: BaseBlockk<P>): boolean;
    /**
     * @param other
     * @return Has this a prependable segment to {@link other }?
     */
    hasPrependable(other: BaseBlockk<P>): boolean;
    /**
     * @param anchor
     * @return A pair that includes the index of `anchor` in the block and
     *  its order relation.
     *  Ordering.EQUAL means that {@link anchor} is realted to a position
     *  of this block.
     */
    indexFrom(anchor: Anchor<P>): u32;
    /**
     * @example
     * a.compare(b) == Ordering.Before if a < b
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    compare(other: BaseBlockk<P>): BlockOrdering;
    /**
     * @example
     * Block(p, "ab").append(Block(_, "cd")) == Block(p, "abcd")
     *
     * @param other
     *      {@link other } is appendable to this.
     * @return other appended to this.
     */
    append(other: Block<P, E>): Block<P, E>;
    /**
     * @example
     * Block(p, "ab").leftSplitAt(1) == Block(p, "a")
     * Block(p, "ab").leftSplitAt(2) == Block(p, "ab")
     *
     * @param index 0-based index
     *      0 < {@link index } <= {@link Block#length }
     * @return Left split.
     */
    leftSplitAt(index: u32): Block<P, E>;
    /**
     * @example
     * Block(p, "ab").rightSplitAt(1) == Block(p+1, "b")
     * Block(p, "ab").rightSplitAt(0) == Block(p, "ab")
     *
     * @param index 0-based index
     *      {@link index } < this.length
     * @return Right split.
     */
    rightSplitAt(index: u32): Block<P, E>;
    /**
     * @example
     * b.splitAt(n) == [b.leftSplitAt(n), b.rightSplitAt(n)]
     *
     * @param index 0-based index
     *      0 < {@link index } < {@link Block#length }
     * @return Left and right splits.
     */
    splitAt(index: u32): readonly [Block<P, E>, Block<P, E>];
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return Left and right splits.
     */
    splitWith(other: BaseBlockk<P>): readonly [Block<P, E>, Block<P, E>];
    /**
     * @param other
     *      this includes or overlaps after {@link other }
     * @return Part of this block which can be append to {@link block }.
     */
    appendable(other: BaseBlockk<P>): Block<P, E>;
    /**
     * @param other
     *      this includes or overlaps before {@link other }
     * @return Part of this block which can be prepend to {@link block }.
     */
    prependable(other: BaseBlockk<P>): Block<P, E>;
    /**
     * @param other
     *      this and {@link other } intersect.
     * @return Intersection part between this and {@link other }.
     */
    intersection(other: BaseBlockk<P>): Block<P, E>;
}
