import { DotPos } from "./dot-pos";
import { Concat } from "./concat";
import { OpReplicatedList, OpEditableReplicatedList } from "./op-replicated-list";
import { Del, Ins } from "./local-operation";
import { LengthBlock, Block } from "./block";
import { u32 } from "../util/number";
import { FromPlain } from "../util/data-validation";
import { Anchor } from "./anchor";
export declare class DeltaReplicatedList<P extends DotPos<P>, E extends Concat<E>> {
    static from<P extends DotPos<P>, E extends Concat<E>>(list: OpReplicatedList<P, E>): DeltaReplicatedList<P, E>;
    /**
     * @param opFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain<P extends DotPos<P>, E extends Concat<E>>(opFromPlain: FromPlain<OpReplicatedList<P, E>>): FromPlain<DeltaReplicatedList<P, E>>;
    protected readonly list: OpReplicatedList<P, E>;
    /**
     * Map replica to their last observed seq.
     */
    protected readonly versionVector: {
        [replica: string]: u32 | undefined;
    };
    protected constructor(list: OpReplicatedList<P, E>, vv: {
        [r: string]: u32;
    });
    readonlyVersionvector(): {
        readonly [replica: string]: u32 | undefined;
    };
    readonly length: u32;
    reduceBlock<U>(f: (acc: U, b: Block<P, E>) => U, prefix: U): U;
    /**
     * @param prefix
     * @return Concatenated version prefixed by `prefix`.
     */
    concatenated(prefix: E): E;
    /**
     * Non-cryptographic way to approximate object identity.
     * Do not take the blocks' content into account.
     */
    structuralHashCode(): u32;
    private lastSeq;
    /**
     * @param anchor
     * @return index of `anchor`.
     */
    indexFrom(anchor: Anchor<P>): u32;
    /**
     * Complexity: O(1)
     *
     * @param delta
     * @return parts of `delta` that can be inserted.
     */
    insertable(delta: Block<P, E>): Block<P, E>[];
    /**
     * Complexity: O(S) where S depends of the chosen implementation.
     * e.g. for a balanced tree: S = log2(N) with N the number of blocks.
     *
     * @param delta
     * @return parts of `delta` that was removed.
     */
    removed(delta: Block<P, E>): LengthBlock<P>[];
    protected bumpVersion(delta: Block<P, E> | LengthBlock<P>): void;
    insert(delta: Block<P, E>): Ins<E>[];
    remove(delta: LengthBlock<P>): Del[];
    quickFilter(delta: Block<P, E>): Block<P, E>[];
    quickFilter(delta: LengthBlock<P>): LengthBlock<P>[];
    quickFilter(delta: LengthBlock<P> | Block<P, E>): (LengthBlock<P> | Block<P, E>)[];
    /**
     * [Mutation]
     *
     * @param delta
     *  if {@link delta } is a {@link Block }, it is an insertion.
     *  if {@link delta } is a {@link LengthBlock }, it is a removal.
     * @return Performed modifications in terms of local operations.
     *  The n+1 -th operation depends on the n -th operation.
     */
    applyDelta(delta: LengthBlock<P>): Del[];
    applyDelta(delta: Block<P, E>): Ins<E>[];
    applyDelta(delta: LengthBlock<P> | Block<P, E>): Del[] | Ins<E>[];
    /**
     * [Mutation]
     * Unidirectional merge from other to this.
     * This operation is idempotent.
     *
     * @param other
     * @return Performed modifications in terms of local operations.
     *  The n+1 -th operation depends on the n -th operation.
     */
    merge(other: DeltaReplicatedList<P, E>): (Ins<E> | Del)[];
}
export declare class DeltaEditableReplicatedList<P extends DotPos<P>, E extends Concat<E>> extends DeltaReplicatedList<P, E> {
    static from<P extends DotPos<P>, E extends Concat<E>>(list: OpEditableReplicatedList<P, E>): DeltaEditableReplicatedList<P, E>;
    /**
     * @param opFromPlain
     * @return function that accepts a value and attempt to build a list.
     *  It returns the built list if it succeeds, or undefined if it fails.
     */
    static fromPlain<P extends DotPos<P>, E extends Concat<E>>(opFromPlain: FromPlain<OpEditableReplicatedList<P, E>>): FromPlain<DeltaEditableReplicatedList<P, E>>;
    protected readonly list: OpEditableReplicatedList<P, E>;
    protected constructor(list: OpEditableReplicatedList<P, E>, vv: {
        [r: string]: u32;
    });
    /**
     * @param index index where the anchor is
     * @param isAfter Is the anchor after `index`?
     * @return anchor at `index`.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    anchorAt(index: u32, isAfter: boolean): Anchor<P>;
    /**
     * [Mutation]
     * Insert {@link items } at {@link index}.
     *
     * @param index 0-based index; Where to insert.
     * @param items elements to insert.
     * @return Delta which represents the insertion.
     */
    insertAt(index: u32, items: E): Block<P, E>;
    /**
     * [Mutation]
     * Remove a number of {@link length } elements from {@link index }
     *
     * @param index 0-based index.
     * @param length Number of elements to remove.
     * @return Delta which represents the deletion.
     */
    removeAt(index: u32, length: u32): LengthBlock<P>[];
}
