import { u32 } from "../util/number";
/**
 * Possible order relation between two intervals.
 *
 * @example
 * [0, 2] < [4, 6]
 * [0, 2] <: [3, 6]
 * [0, 4] <∩ [4, 6] (because their intersection is not empty)
 * [0, 2] -⊂ [0, 3]
 * [1, 2] ⊂ [0, 3]
 * [2, 3] +⊂ [0, 3]
 * [0, 2] = [0, 2]
 */
export declare const enum RangeOrdering {
    BEFORE = -6,
    PREPENDABLE = -5,
    OVERLAPPING_BEFORE = -4,
    INCLUDED_LEFT_BY = -3,
    INCLUDED_MIDDLE_BY = -2,
    INCLUDED_RIGHT_BY = -1,
    EQUAL = 0,
    INCLUDING_LEFT = 1,
    INCLUDING_MIDDLE = 2,
    INCLUDING_RIGHT = 3,
    OVERLAPPING_AFTER = 4,
    APPENDABLE = 5,
    AFTER = 6
}
/**
 * Interval of u32.
 */
export declare class U32Range {
    readonly lower: u32;
    readonly length: u32;
    /**
     * @param lower Lower bound.
     * @param length Number of elements in the range.
     */
    protected constructor(lower: u32, length: u32);
    /**
     * @param lower {@link IntInterval#lower }
     * @param length {@link IntInterval#length }
     * @return New range.
     */
    static fromLength(lower: u32, length: u32): U32Range;
    /**
     * @param lower {@link IntInterval#lower }
     * @param upper {@link IntInterval#length }
     * @return New range.
     */
    static fromBounds(lower: u32, upper: u32): U32Range;
    /**
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    static fromPlain(x: unknown): U32Range | undefined;
    /**
     * @example
     * nth(0) == lower
     * nth(length - 1) == upper
     *
     * @param nth 0-based
     * @return n-th element.
     */
    nth(nth: u32): u32;
    /**
     * Upper bound.
     */
    upper(): u32;
    /**
     * @param other
     * @return {@link other } appended to this.
     */
    append(other: U32Range): U32Range;
    /**
     * @param other
     *      this includes or overlaps after {@link other }
     * @return Part of this block which can be append to {@link block }.
     */
    appendable(other: U32Range): U32Range;
    /**
     * @param other
     *      this includes or overlaps before {@link other }
     * @return Part of this block which can be prepend to {@link block }.
     */
    prependable(other: U32Range): U32Range;
    /**
     * @param other
     *      this and {@link other } intersect.
     * @return Intersection part between this and {@link other }.
     */
    intersection(other: U32Range): U32Range;
    /**
     * @param other
     * @return Are this and {@link other } an intersection?
     */
    hasIntersection(other: U32Range): boolean;
    /**
     * @param other
     * @return Has this an appendable segment to {@link other }?
     */
    hasAppendable(other: U32Range): boolean;
    /**
     * @param other
     * @return Has this a prependable segment to {@link other }?
     */
    hasPrependable(other: U32Range): boolean;
    /**
     * @param other
     * @return this [order relation] other.
     */
    compare(other: U32Range): RangeOrdering;
}
