import { u32 } from "../../util/number";
import { Ordering } from "../../util/ordering";
/**
 * @final this class cannot be extended
 *
 * Part of a {@link SimpleDotPos}.
 *
 * A part is a lexicographic triplet (priority, replica, seq).
 * The dot (replica, seq) uniquely identifies.
 */
export declare class SimpleDotPosPart {
    /**
     * @param priority {@link SimpleDotPosPart#priority }
     * @param replica {@link SimpleDotPosPart#replica }
     * @param seq {@link SimpleDotPosPart#seq }
     */
    private constructor();
    /**
     * @param priority {@link SimpleDotPosPart#priority }
     *      ∉ {U32_BOTTOM, U32_TOP}
     * @param replica {@link SimpleDotPosPart#replica }
     *      !== U32_TOP
     * @param seq {@link SimpleDotPosPart#seq }
     */
    static from(priority: u32, replica: u32, seq: u32): SimpleDotPosPart;
    /**
     * @note {@link SimpleDotPosPart#BOTTOM} and {@link SimpleDotPosPart#TOP}
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    static fromPlain(x: unknown): SimpleDotPosPart | undefined;
    /**
     * Lowest part.
     * Here the pair (replica, seq) has no meaning.
     */
    static readonly BOTTOM: SimpleDotPosPart;
    /**
     * Greatest part.
     * Here the pair (replica, seq) has no meaning.
     */
    static readonly TOP: SimpleDotPosPart;
    /**
     * @param seq The seq of the new SimpleDotPosPart
     * @return part with the same base,
     *  but with a different seq
     */
    withSeq(seq: u32): SimpleDotPosPart;
    /**
     * Priority of the part over another one.
     * If two parts have the same priority, the lexicographic pair
     * (replica, seq) is used as disambiguator.
     */
    readonly priority: u32;
    /**
     * Globally unique identifier of the author which generated this part.
     */
    readonly replica: u32;
    /**
     * Locally unique sequence number.
     * Each generated part has a sequence number.
     */
    readonly seq: u32;
    /**
     * @return [priority, replica, seq]
     */
    asTuple(): readonly [u32, u32, u32];
    /**
     * Non-cryptographic way to approximate object identity.
     */
    hashCode(): u32;
    /**
     * @example
     * SimpleDotPosPart(0, 0, _) == SimpleDotPosPart(0, 0, _)
     * SimpleDotPosPart(0, 0, _) < SimpleDotPosPart(1, 0, _)
     * SimpleDotPosPart(1, 0, _) < SimpleDotPosPart(1, 1, _)
     *
     * Lexicographic order of pair (priority, replica).
     *
     * @param other
     * @return this [Order relation] {@link other }, regardless their seqs
     *  ({@link SimpleDotPosPart#seq}).
     */
    compareBase(other: SimpleDotPosPart): Ordering;
    /**
     * @example
     * SimpleDotPosPart(0, 0, 0) < SimpleDotPosPart(1, 0, 0)
     * SimpleDotPosPart(1, 0, 0) < SimpleDotPosPart(1, 1, 0)
     * SimpleDotPosPart(1, 1, 0) < SimpleDotPosPart(1, 1, 1)
     *
     * Lexicographic order of triple (priority, replica, seq).
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    compare(other: SimpleDotPosPart): Ordering;
}
