import { u32 } from "../../util/number";
import { BaseOrdering } from "../../core/pos";
import { SimpleDotPosPart } from "./simple-dot-pos-part";
import { Ordering } from "../../util/ordering";
import { DotPos } from "../../core/dot-pos";
/**
 * @final this class cannot be extended
 *
 * A position can be represented as a lexicographic list of
 * lexicographic triplets (priority, replica, seq).
 * Theses triplets are the parts of the position.
 *
 * The dot (replica, seq) uniquely identifies the position.
 *
 * The set of positions is dense, hence you can always author a new position
 * between two distinct positions. A position can be generated between two
 * int-successive positions (see {@link Pos}) by suffixing a new triplet
 * to the lower position.
 */
export declare class SimpleDotPos implements DotPos<SimpleDotPos> {
    /**
     * @param parts {@link SimpleDotPos#parts }
     */
    private constructor();
    /**
     * @param parts {@link SimpleDotPos#parts }
     *  The last part must be distinct of SimplePosPart.BOTTOM and
     *  SimplePosPart.TOP.
     * @return Position with {@link parts } as {@link SimpleDotPos#parts }.
     */
    static from(parts: readonly SimpleDotPosPart[]): SimpleDotPos;
    /**
     * @note {@link SimpleDotPos#BOTTOM } and {@link SimpleDotPos#TOP }
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    static fromPlain(x: unknown): SimpleDotPos | undefined;
    /**
     * Lowest position.
     */
    static readonly BOTTOM: SimpleDotPos;
    /**
     * Greatest position.
     */
    static readonly TOP: SimpleDotPos;
    /**
     * @param seq The seq of the new position
     * @return Position with the same base, but with a different seq.
     */
    withSeq(seq: u32): SimpleDotPos;
    /** @override */
    intSucc(n: u32): SimpleDotPos;
    /**
     * Parts of this position.
     */
    readonly parts: readonly SimpleDotPosPart[];
    /**
     * Last part of {@link SimplePosPart#parts }.
     */
    lastPart(): SimpleDotPosPart;
    /**
     * Number of part in {@link SimplePosPart#parts }.
     */
    depth(): u32;
    /** @override */
    replica(): u32;
    /** @override */
    seq(): u32;
    /** @override */
    intDistance(other: SimpleDotPos): readonly [u32, Ordering];
    /** @override */
    hashCode(): u32;
    /** @override */
    hasIntSucc(n: u32): boolean;
    /** @override */
    compareBase(other: SimpleDotPos): BaseOrdering;
    /** @override */
    isBaseEqual(other: SimpleDotPos): boolean;
    /** @override */
    compare(other: SimpleDotPos): Ordering;
    /** @override */
    isEqual(other: SimpleDotPos): boolean;
}
