import { Block, LengthBlock } from "../../core/block";
import { BlockFactory } from "../../core/block-factory";
import { Concat } from "../../core/concat";
import { Pos } from "../../core/pos";
import { Ins, Del } from "../../core/local-operation";
import { u32 } from "../../util/number";
import { FromPlain } from "../../util/data-validation";
import { Anchor } from "../../core/anchor";
export declare abstract class Linkable<P extends Pos<P>, E extends Concat<E>> {
    /**
     * Next cell
     */
    right?: Cell<P, E>;
    /**
     * @param right Next cell.
     */
    constructor(right?: Cell<P, E>);
    /**
     * @param f reducer that respectively accepts the accumulated value and
     *  the current block as first and second parameters. It returns the next
     *  accumulated value.
     * @param prefix initial value
     * @return accumulated value by reducing blocks from left to right
     *  in the list.
     */
    reduceBlock<U>(f: (acc: U, b: Block<P, E>) => U, prefix: U): U;
    /**
     * @param anchor
     * @param minIndex minimal index in the current sub-list.
     * @return index of `anchor`.
     */
    indexFrom(anchor: Anchor<P>, minIndex: u32): u32;
    /**
     * @param relIndex relative index where the anchor is
     * @param isAfter Is the anchor after or before `relIndex`?
     * @param f factory
     * @return anchor at `index` in the current list.
     *  The anchor is sticked to the left psoition if isAfter is false.
     * Otherwise, it is sticked to the right psoition.
     */
    anchorAt(index: u32, isAfter: boolean, f: BlockFactory<P>): Anchor<P>;
    /**
     * @param iBlock
     * @return segments of `iBlock` that can be inserted in the current
     *  sub-list.
     */
    insertable(iBlock: Block<P, E>): Block<P, E>[];
    /**
     * [Mutation]
     * Insert {@link rblock } to right and preserve the chain.
     *
     * @param rblock block to insert.
     * @return inserted cell with {@link rblock } as content.
     */
    protected insertRight(rblock: Block<P, E>): Cell<P, E>;
    /**
     * [Mutation]
     * Insert the parts of {@link iBlock } which are not already inserted.
     *
     * @param iBlock block to insert.
     * @param index 0-based index
     * @return Performed modifications in terms of local insertions.
     *  The n+1 -th insertion depends on the effect of the n -th insertion.
     */
    insert(iBlock: Block<P, E>, index: u32): Ins<E>[];
    /**
     * [Mutation]
     * Insert {@link items } at {@link index}.
     *
     * @param index 0-based index; Where to insert.
     * @param items elements to insert.
     * @param factory strategy of block generation.
     * @return Delta which represents the insertion.
     */
    insertAt(index: u32, items: E, factory: BlockFactory<P>): Block<P, E>;
    /**
     * [Mutation]
     * Remove a number of {@link length } elements from {@link index }
     *
     * @param index 0-based index.
     * @param length Number of elements to remove.
     * @return Delta which represents the deletion.
     */
    removeAt(index: u32, length: u32): LengthBlock<P>[];
    /**
     * [Mutation]
     * Remove {@link dBlock }.
     * @param dBlock block to remove.
     * @param index index of the current block in the chain.
     * @return Performed modifications in terms of local deletions.
     *  The n+1 -th deletion depends on the n -th deletion.
     */
    remove(dBlock: LengthBlock<P> | Block<P, E>, index: u32): Del[];
}
export declare class Sentinel<P extends Pos<P>, E extends Concat<E>> extends Linkable<P, E> {
    /**
     * @param blockFromPlain
     * @return function that accepts a value and returns a Sentinel from
     *  the value, or undefined if the value is mal-formed
     */
    static fromPlain<P extends Pos<P>, E extends Concat<E>>(blockFromPlain: FromPlain<Block<P, E>>): FromPlain<Sentinel<P, E>>;
}
export declare class Cell<P extends Pos<P>, E extends Concat<E>> extends Linkable<P, E> {
    /**
     * @param blockFromPlain
     * @return function that accepts a value and returns a Cell from
     *  the value, or undefined if the value is mal-formed
     */
    static fromPlain<P extends Pos<P>, E extends Concat<E>>(blockFromPlain: FromPlain<Block<P, E>>): FromPlain<Cell<P, E>>;
    /**
     * @param block {@link Cell#block }
     * @param right {@link Linkable#right }
     */
    constructor(block: Block<P, E>, right: Cell<P, E> | undefined);
    /**
     * Block attached to this cell.
     */
    readonly block: Block<P, E>;
}
