import { u32 } from "../../util/number";
import { Ordering } from "../../util/ordering";
/**
 * @final this class cannot be extended
 *
 * Part of a {@link SimplePos}.
 *
 * A part is a lexicographic quadruple (priority, replica, nth, offset).
 * The triplet (replica, nth, offset) uniquely identifies a part.
 */
export declare class SimplePosPart {
    /**
     * Lowest part.
     * Here the triplet (replica, nth, offset) has no meaning.
     */
    static readonly BOTTOM: SimplePosPart;
    /**
     * Greatest part.
     * Here the triplet (replica, nth, offset) has no meaning.
     */
    static readonly TOP: SimplePosPart;
    /**
     * @param priority {@link SimplePosPart#priority }
     *      ∉ {U32_BOTTOM, U32_TOP}
     * @param replica {@link SimplePosPart#replica }
     *      !== U32_TOP
     * @param nth {@link SimplePosPart#nth }
     * @param offset {@link SimplePosPart#offset }
     */
    static from(priority: u32, replica: u32, nth: u32, offset: u32): SimplePosPart;
    /**
     * @note {@link SimplePosPart#BOTTOM} and {@link SimplePosPart#TOP}
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x`, or undefined if `x` is not valid.
     */
    static fromPlain(x: unknown): SimplePosPart | undefined;
    /**
     * Priority of the part over another one.
     * If two parts have the same priority, the lexicographic pair
     * (replica, nth) is used as disambiguator.
     */
    readonly priority: u32;
    /**
     * Globally unique identifier of the author which generated this part.
     */
    readonly replica: u32;
    /**
     * Each part is generated for the `nth` generated block.
     */
    readonly nth: u32;
    /**
     * Offset in the block.
     */
    readonly offset: u32;
    /**
     * @param priority {@link SimplePosPart#priority }
     * @param replica {@link SimplePosPart#replica }
     * @param nth {@link SimplePosPart#nth }
     * @param offset {@link SimplePosPart#offset }
     */
    private constructor();
    /**
     * @param offset The offset of the new SimplePosPart
     * @return part with the same base, but with a different offset
     */
    withOffset(offset: u32): SimplePosPart;
    /**
     * @return [priority, replica, nth, offset]
     */
    asTuple(): readonly [u32, u32, u32, u32];
    /**
     * Non-cryptographic way to approximate object identity.
     */
    hashCode(): u32;
    /**
     * @example
     * SimplePosPart(0, 0, 0, _) == SimplePosPart(0, 0, 0, _)
     * SimplePosPart(0, 0, 0, _) < SimplePosPart(1, 0, 0, _)
     * SimplePosPart(1, 0, 0, _) < SimplePosPart(1, 1, 0, _)
     * SimplePosPart(1, 1, 0, _) < SimplePosPart(1, 1, 1, _)
     *
     * Lexicographic order of triplet (priority, replica, nth).
     *
     * @param other
     * @return `this` [Order relation] `other` regardless their offsets
     *  ({@link SimplePosPart#offset})
     */
    compareBase(other: SimplePosPart): Ordering;
    /**
     * @example
     * SimplePosPart(0, 0, 0, 0) < SimplePosPart(1, 0, 0, 0)
     * SimplePosPart(1, 0, 0, 0) < SimplePosPart(1, 1, 0, 0)
     * SimplePosPart(1, 1, 0, 0) < SimplePosPart(1, 1, 1, 0)
     * SimplePosPart(1, 1, 1, 0) < SimplePosPart(1, 1, 1, 1)
     *
     * Lexicographic order of quadruple (priority, replica, nth, offset).
     *
     * @param other
     * @return `this` [Order relation] `other`
     */
    compare(other: SimplePosPart): Ordering;
}
