import { u32 } from "../../util/number";
import { BaseOrdering } from "../../core/pos";
import { SimplePosPart } from "./simple-pos-part";
import { Ordering } from "../../util/ordering";
import { Pos } from "../../core/pos";
/**
 * @final this class cannot be extended
 *
 * A position can be represented as a lexicographic list of
 * lexicographic quadruples (priority, replica, nth, offset).
 * Theses triplets are the parts of the position.
 *
 * The triplet (replica, nth, offset) uniquely identifies the position.
 *
 * The set of positions is dense, hence you can always author a new position
 * between two distinct positions. A position can be generated between two
 * int-successive positions (see {@link Pos}) by suffixing a new triplet
 * to the lower position.
 */
export declare class SimplePos implements Pos<SimplePos> {
    /**
     * Lowest position.
     */
    static readonly BOTTOM: SimplePos;
    /**
     * Greatest position.
     */
    static readonly TOP: SimplePos;
    /**
     * @param parts {@link SimplePos#parts }
     *  The last part must be distinct of SimplePosPart.BOTTOM and
     *  SimplePosPart.TOP.
     * @return Position with {@link parts } as {@link SimplePos#parts }.
     */
    static from(parts: readonly SimplePosPart[]): SimplePos;
    /**
     * @note {@link SimplePos#BOTTOM } and {@link SimplePos#TOP }
     * are not valid candidates.
     *
     * @param x candidate
     * @return object from `x', or undefined if `x' is not valid.
     */
    static fromPlain(x: unknown): SimplePos | undefined;
    /**
     * Parts of this position.
     */
    readonly parts: readonly SimplePosPart[];
    /**
     * @param parts {@link SimplePos#parts }
     */
    private constructor();
    /**
     * @param offset The offset of the new position
     * @return Position with the same base, but with a different offset
     */
    withOffset(offset: u32): SimplePos;
    /**
     * @param n 0-based index.
     * @return `n`-th integer predecessor of this.
     */
    intPred(n: u32): SimplePos;
    /** @override */
    intSucc(n: u32): SimplePos;
    /**
     * Last part of {@link SimplePosPart#parts }.
     */
    lastPart(): SimplePosPart;
    /**
     * Number of part in {@link SimplePosPart#parts }.
     */
    depth(): u32;
    /** @override */
    replica(): u32;
    /**
     * Each position is generated for the `nth` generated block.
     * The pair (replica, nth) globally and uniquely identifies a block.
     */
    nth(): u32;
    /**
     * Offset of the position in the block.
     */
    offset(): u32;
    /** @override */
    intDistance(other: SimplePos): readonly [u32, Ordering];
    /** @override */
    hashCode(): u32;
    /**
     * hasIntPred(0) is always true.
     *
     * See also {@link Pos#hasIntSucc }.
     *
     * @param n 0-based index.
     * @return is there a {@link n } -th integer predecessor?
     */
    hasIntPred(n: u32): boolean;
    /** @override */
    hasIntSucc(n: u32): boolean;
    /** @override */
    compareBase(other: SimplePos): BaseOrdering;
    /** @override */
    isBaseEqual(other: SimplePos): boolean;
    /** @override */
    compare(other: SimplePos): Ordering;
    /** @override */
    isEqual(other: SimplePos): boolean;
}
